import re

import pytest
from sqlalchemy import Column
from sqlalchemy import MetaData
from sqlalchemy import Table
from sqlalchemy.sql import func
from sqlalchemy.sql import insert
from sqlalchemy.sql import text

from geoalchemy2.exc import ArgumentError
from geoalchemy2.types import Geography
from geoalchemy2.types import Geometry
from geoalchemy2.types import Raster

from . import select


def eq_sql(a, b):
    a = re.sub(r'[\n\t]', '', str(a))
    assert a == b


@pytest.fixture
def geometry_table():
    table = Table('table', MetaData(), Column('geom', Geometry))
    return table


@pytest.fixture
def geography_table():
    table = Table('table', MetaData(), Column('geom', Geography))
    return table


@pytest.fixture
def raster_table():
    table = Table('table', MetaData(), Column('rast', Raster))
    return table


class TestGeometry():

    def test_get_col_spec(self):
        g = Geometry(srid=900913)
        assert g.get_col_spec() == 'geometry(GEOMETRY,900913)'

    def test_get_col_spec_no_typmod(self):
        g = Geometry(geometry_type=None)
        assert g.get_col_spec() == 'geometry'

    def test_check_ctor_args_bad_srid(self):
        with pytest.raises(ArgumentError):
            Geometry(srid='foo')

    def test_get_col_spec_geometryzm(self):
        g = Geometry(geometry_type='GEOMETRYZM', srid=900913)
        assert g.get_col_spec() == 'geometry(GEOMETRYZM,900913)'

    def test_get_col_spec_geometryz(self):
        g = Geometry(geometry_type='GEOMETRYZ', srid=900913)
        assert g.get_col_spec() == 'geometry(GEOMETRYZ,900913)'

    def test_get_col_spec_geometrym(self):
        g = Geometry(geometry_type='GEOMETRYM', srid=900913)
        assert g.get_col_spec() == 'geometry(GEOMETRYM,900913)'

    def test_check_ctor_args_management_zm(self):
        with pytest.raises(ArgumentError):
            Geometry(geometry_type='POINTZM', management=True)

    def test_check_ctor_args_management_z(self):
        with pytest.raises(ArgumentError):
            Geometry(geometry_type='POINTZ', dimension=2, management=True)

    def test_check_ctor_args_management_m(self):
        with pytest.raises(ArgumentError):
            Geometry(geometry_type='POINTM', dimension=2, management=True)

    def test_check_ctor_args_incompatible_arguments(self):
        with pytest.raises(ArgumentError):
            Geometry(geometry_type=None, management=True)

    def test_check_ctor_args_srid_not_enforced(self):
        with pytest.warns(UserWarning):
            Geometry(geometry_type=None, srid=4326)

    def test_check_ctor_args_use_typmod_ignored(self):
        with pytest.warns(UserWarning):
            Geometry(management=False, use_typmod=True)

    def test_check_ctor_args_use_typmod_nullable(self):
        with pytest.warns(UserWarning, match="use_typmod ignored when management is False"):
            with pytest.raises(
                ArgumentError,
                match='The "nullable" and "use_typmod" arguments can not be used together',
            ):
                Geometry(use_typmod=True, nullable=False)

    def test_column_expression(self, geometry_table):
        s = select([geometry_table.c.geom])
        eq_sql(s, 'SELECT ST_AsEWKB("table".geom) AS geom FROM "table"')

    def test_select_bind_expression(self, geometry_table):
        s = select([text('foo')]).where(geometry_table.c.geom == 'POINT(1 2)')
        eq_sql(s, 'SELECT foo FROM "table" WHERE '
                  '"table".geom = ST_GeomFromEWKT(:geom_1)')
        assert s.compile().params == {'geom_1': 'POINT(1 2)'}

    def test_insert_bind_expression(self, geometry_table):
        i = insert(geometry_table).values(geom='POINT(1 2)')
        eq_sql(i, 'INSERT INTO "table" (geom) VALUES (ST_GeomFromEWKT(:geom))')
        assert i.compile().params == {'geom': 'POINT(1 2)'}

    def test_function_call(self, geometry_table):
        s = select([geometry_table.c.geom.ST_Buffer(2)])
        eq_sql(s,
               'SELECT ST_AsEWKB(ST_Buffer("table".geom, :ST_Buffer_2)) '
               'AS "ST_Buffer_1" FROM "table"')

    def test_non_ST_function_call(self, geometry_table):
        with pytest.raises(AttributeError):
            geometry_table.c.geom.Buffer(2)

    def test_subquery(self, geometry_table):
        # test for geometry columns not delivered to the result
        # http://hg.sqlalchemy.org/sqlalchemy/rev/f1efb20c6d61
        s = select([geometry_table]).alias('name').select()
        eq_sql(s,
               'SELECT ST_AsEWKB(name.geom) AS geom FROM '
               '(SELECT "table".geom AS geom FROM "table") AS name')


class TestGeography():

    def test_get_col_spec(self):
        g = Geography(srid=900913)
        assert g.get_col_spec() == 'geography(GEOMETRY,900913)'

    def test_get_col_spec_no_typmod(self):
        g = Geography(geometry_type=None)
        assert g.get_col_spec() == 'geography'

    def test_column_expression(self, geography_table):
        s = select([geography_table.c.geom])
        eq_sql(s, 'SELECT ST_AsBinary("table".geom) AS geom FROM "table"')

    def test_select_bind_expression(self, geography_table):
        s = select([text('foo')]).where(geography_table.c.geom == 'POINT(1 2)')
        eq_sql(s, 'SELECT foo FROM "table" WHERE '
                  '"table".geom = ST_GeogFromText(:geom_1)')
        assert s.compile().params == {'geom_1': 'POINT(1 2)'}

    def test_insert_bind_expression(self, geography_table):
        i = insert(geography_table).values(geom='POINT(1 2)')
        eq_sql(i, 'INSERT INTO "table" (geom) VALUES (ST_GeogFromText(:geom))')
        assert i.compile().params == {'geom': 'POINT(1 2)'}

    def test_function_call(self, geography_table):
        s = select([geography_table.c.geom.ST_Buffer(2)])
        eq_sql(s,
               'SELECT ST_AsEWKB(ST_Buffer("table".geom, :ST_Buffer_2)) '
               'AS "ST_Buffer_1" FROM "table"')

    def test_non_ST_function_call(self, geography_table):
        with pytest.raises(AttributeError):
            geography_table.c.geom.Buffer(2)

    def test_subquery(self, geography_table):
        # test for geography columns not delivered to the result
        # http://hg.sqlalchemy.org/sqlalchemy/rev/f1efb20c6d61
        s = select([geography_table]).alias('name').select()
        eq_sql(s,
               'SELECT ST_AsBinary(name.geom) AS geom FROM '
               '(SELECT "table".geom AS geom FROM "table") AS name')


class TestPoint():

    def test_get_col_spec(self):
        g = Geometry(geometry_type='POINT', srid=900913)
        assert g.get_col_spec() == 'geometry(POINT,900913)'


class TestCurve():

    def test_get_col_spec(self):
        g = Geometry(geometry_type='CURVE', srid=900913)
        assert g.get_col_spec() == 'geometry(CURVE,900913)'


class TestLineString():

    def test_get_col_spec(self):
        g = Geometry(geometry_type='LINESTRING', srid=900913)
        assert g.get_col_spec() == 'geometry(LINESTRING,900913)'


class TestPolygon():

    def test_get_col_spec(self):
        g = Geometry(geometry_type='POLYGON', srid=900913)
        assert g.get_col_spec() == 'geometry(POLYGON,900913)'


class TestMultiPoint():

    def test_get_col_spec(self):
        g = Geometry(geometry_type='MULTIPOINT', srid=900913)
        assert g.get_col_spec() == 'geometry(MULTIPOINT,900913)'


class TestMultiLineString():

    def test_get_col_spec(self):
        g = Geometry(geometry_type='MULTILINESTRING', srid=900913)
        assert g.get_col_spec() == 'geometry(MULTILINESTRING,900913)'


class TestMultiPolygon():

    def test_get_col_spec(self):
        g = Geometry(geometry_type='MULTIPOLYGON', srid=900913)
        assert g.get_col_spec() == 'geometry(MULTIPOLYGON,900913)'


class TestGeometryCollection():

    def test_get_col_spec(self):
        g = Geometry(geometry_type='GEOMETRYCOLLECTION', srid=900913)
        assert g.get_col_spec() == 'geometry(GEOMETRYCOLLECTION,900913)'


class TestRaster():

    def test_get_col_spec(self):
        r = Raster()
        assert r.get_col_spec() == 'raster'

    def test_column_expression(self, raster_table):
        s = select([raster_table.c.rast])
        eq_sql(s, 'SELECT raster("table".rast) AS rast FROM "table"')

    def test_insert_bind_expression(self, raster_table):
        i = insert(raster_table).values(rast=b'\x01\x02')
        eq_sql(i, 'INSERT INTO "table" (rast) VALUES (raster(:rast))')
        assert i.compile().params == {'rast': b'\x01\x02'}

    def test_function_call(self, raster_table):
        s = select([raster_table.c.rast.ST_Height()])
        eq_sql(s,
               'SELECT ST_Height("table".rast) '
               'AS "ST_Height_1" FROM "table"')

    def test_non_ST_function_call(self, raster_table):

        with pytest.raises(AttributeError):
            raster_table.c.geom.Height()


class TestCompositeType():

    def test_ST_Dump(self, geography_table):
        s = select([func.ST_Dump(geography_table.c.geom).geom.label("geom")])

        eq_sql(s,
               'SELECT ST_AsEWKB((ST_Dump("table".geom)).geom) AS geom '
               'FROM "table"')



"""Declare tables used in tests."""
import pytest
from sqlalchemy import Column
from sqlalchemy import Integer
from sqlalchemy import MetaData
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.sql import func
from sqlalchemy.types import TypeDecorator

from geoalchemy2 import Geography
from geoalchemy2 import Geometry
from geoalchemy2 import Raster


@pytest.fixture
def Lake(base, postgis_version, schema):
    with_management = postgis_version.startswith("1.")

    class Lake(base):
        __tablename__ = 'lake'
        __table_args__ = {'schema': schema}
        id = Column(Integer, primary_key=True)
        geom = Column(Geometry(geometry_type='LINESTRING', srid=4326, management=with_management))

        def __init__(self, geom):
            self.geom = geom

    return Lake


@pytest.fixture
def Poi(base, engine, schema):
    class Poi(base):
        __tablename__ = 'poi'
        __table_args__ = {'schema': schema}
        id = Column(Integer, primary_key=True)
        geom = Column(Geometry(geometry_type='POINT', srid=4326))
        geog = (
            Column(Geography(geometry_type='POINT', srid=4326))
            if engine.dialect.name == "postgresql"
            else None
        )

        def __init__(self, geog):
            self.geog = geog

    return Poi


@pytest.fixture
def Summit(base, postgis_version, schema):
    with_use_typemod = postgis_version.startswith("1.")

    class Summit(base):
        __tablename__ = 'summit'
        __table_args__ = {'schema': schema}
        id = Column(Integer, primary_key=True)
        geom = Column(Geometry(
            geometry_type='POINT', srid=4326, management=True, use_typmod=with_use_typemod))

        def __init__(self, geom):
            self.geom = geom

    return Summit


@pytest.fixture
def Ocean(base, postgis_version):
    # The raster type is only available on PostGIS 2.0 and above
    if postgis_version.startswith('1.'):
        pytest.skip("The raster type is only available on PostGIS 2.0 and above")

    class Ocean(base):
        __tablename__ = 'ocean'
        id = Column(Integer, primary_key=True)
        rast = Column(Raster)

        def __init__(self, rast):
            self.rast = rast

    return Ocean


class ThreeDGeometry(TypeDecorator):
    """This class is used to insert a ST_Force3D() in each insert."""
    impl = Geometry

    def bind_expression(self, bindvalue):
        return func.ST_Force3D(self.impl.bind_expression(bindvalue))


@pytest.fixture
def PointZ(base):
    class PointZ(base):
        __tablename__ = "point_z"
        id = Column(Integer, primary_key=True)
        three_d_geom = Column(ThreeDGeometry(srid=4326, geometry_type="POINTZ", dimension=3))

    return PointZ


class TransformedGeometry(TypeDecorator):
    """This class is used to insert a ST_Transform() in each insert or select."""
    impl = Geometry

    def __init__(self, db_srid, app_srid, **kwargs):
        kwargs["srid"] = db_srid
        self.impl = self.__class__.impl(**kwargs)
        self.app_srid = app_srid
        self.db_srid = db_srid

    def column_expression(self, col):
        """The column_expression() method is overridden to ensure that the
        SRID of the resulting WKBElement is correct"""
        return getattr(func, self.impl.as_binary)(
            func.ST_Transform(col, self.app_srid),
            type_=self.__class__.impl(srid=self.app_srid)
            # srid could also be -1 so that the SRID is deduced from the
            # WKB data
        )

    def bind_expression(self, bindvalue):
        return func.ST_Transform(
            self.impl.bind_expression(bindvalue), self.db_srid)


@pytest.fixture
def LocalPoint(base):
    class LocalPoint(base):
        __tablename__ = "local_point"
        id = Column(Integer, primary_key=True)
        geom = Column(
            TransformedGeometry(
                db_srid=2154, app_srid=4326, geometry_type="POINT", management=False))
        managed_geom = Column(
            TransformedGeometry(
                db_srid=2154, app_srid=4326, geometry_type="POINT", management=True))

    return LocalPoint


@pytest.fixture
def IndexTestWithSchema(base, schema):
    class IndexTestWithSchema(base):
        __tablename__ = 'indextestwithschema'
        __table_args__ = {'schema': schema} if schema else {}
        id = Column(Integer, primary_key=True)
        geom1 = Column(Geometry(geometry_type='POINT', srid=4326))
        geom2 = Column(Geometry(geometry_type='POINT', srid=4326, management=True))

    return IndexTestWithSchema


@pytest.fixture
def IndexTestWithNDIndex(base, schema):
    class IndexTestWithNDIndex(base):
        __tablename__ = 'index_test_with_nd_index'
        __table_args__ = {'schema': schema}
        id = Column(Integer, primary_key=True)
        geom1 = Column(Geometry(geometry_type='POINTZ', dimension=3, use_N_D_index=True))

    return IndexTestWithNDIndex


@pytest.fixture
def IndexTestWithoutSchema(base):
    class IndexTestWithoutSchema(base):
        __tablename__ = 'indextestwithoutschema'
        id = Column(Integer, primary_key=True)
        geom1 = Column(Geometry(geometry_type='POINT', srid=4326))
        geom2 = Column(Geometry(geometry_type='POINT', srid=4326, management=True))

    return IndexTestWithoutSchema


@pytest.fixture
def reflection_tables_metadata():
    metadata = MetaData()
    base = declarative_base(metadata=metadata)

    class Lake(base):
        __tablename__ = 'lake'
        id = Column(Integer, primary_key=True)
        geom = Column(Geometry(geometry_type='LINESTRING', srid=4326))
        geom_no_idx = Column(Geometry(geometry_type='LINESTRING', srid=4326, spatial_index=False))
        geom_z = Column(Geometry(geometry_type='LINESTRINGZ', srid=4326, dimension=3))
        geom_m = Column(Geometry(geometry_type='LINESTRINGM', srid=4326, dimension=3))
        geom_zm = Column(Geometry(geometry_type='LINESTRINGZM', srid=4326, dimension=4))

    return metadata



import pytest
from sqlalchemy import Column
from sqlalchemy import Integer
from sqlalchemy import MetaData
from sqlalchemy.ext.declarative import declarative_base

from geoalchemy2 import Geometry
from geoalchemy2 import WKBElement
from geoalchemy2 import WKTElement

metadata = MetaData()
Base = declarative_base(metadata=metadata)


class PickledLake(Base):
    __tablename__ = 'pickled_lake'
    id = Column(Integer, primary_key=True)
    geom = Column(Geometry(geometry_type='LINESTRING', srid=4326))

    def __init__(self, geom):
        self.geom = geom


class TestPickle():

    @pytest.fixture
    def setup_one_lake(self, session):
        conn = session.bind
        metadata.drop_all(conn, checkfirst=True)
        metadata.create_all(conn)

        lake = PickledLake(WKTElement('LINESTRING(0 0,1 1)', srid=4326))
        session.add(lake)
        session.flush()
        session.expire(lake)

        yield lake.id

        session.rollback()
        metadata.drop_all(session.bind, checkfirst=True)

    def test_pickle_unpickle(self, session, setup_one_lake):
        import pickle

        lake_id = setup_one_lake

        lake = session.query(PickledLake).get(lake_id)
        assert isinstance(lake.geom, WKBElement)
        data_desc = str(lake.geom)

        pickled = pickle.dumps(lake)
        unpickled = pickle.loads(pickled)
        assert unpickled.geom.srid == 4326
        assert str(unpickled.geom) == data_desc
        assert unpickled.geom.extended is True



""" This module defines the :class:`geoalchemy2.types.Geometry`,
:class:`geoalchemy2.types.Geography`, and :class:`geoalchemy2.types.Raster`
classes, that are used when defining geometry, geography and raster
columns/properties in models.

Reference
---------
"""
import warnings

from sqlalchemy.dialects import postgresql
from sqlalchemy.dialects.postgresql.base import ischema_names as postgresql_ischema_names
from sqlalchemy.dialects.sqlite.base import ischema_names as sqlite_ischema_names
from sqlalchemy.sql import func
from sqlalchemy.types import Float
from sqlalchemy.types import Integer
from sqlalchemy.types import UserDefinedType

try:
    from .shape import to_shape
    SHAPELY = True
except ImportError:
    SHAPELY = False


from .comparator import BaseComparator
from .comparator import Comparator
from .elements import CompositeElement
from .elements import RasterElement
from .elements import WKBElement
from .elements import WKTElement
from .exc import ArgumentError


class _GISType(UserDefinedType):
    """
    The base class for :class:`geoalchemy2.types.Geometry` and
    :class:`geoalchemy2.types.Geography`.

    This class defines ``bind_expression`` and ``column_expression`` methods
    that wrap column expressions in ``ST_GeomFromEWKT``, ``ST_GeogFromText``,
    or ``ST_AsEWKB`` calls.

    This class also defines ``result_processor`` and ``bind_processor``
    methods. The function returned by ``result_processor`` converts WKB values
    received from the database to :class:`geoalchemy2.elements.WKBElement`
    objects. The function returned by ``bind_processor`` converts
    :class:`geoalchemy2.elements.WKTElement` objects to EWKT strings.

    Args:
        geometry_type: The geometry type.

            Possible values are:

              * ``"GEOMETRY"``,
              * ``"POINT"``,
              * ``"LINESTRING"``,
              * ``"POLYGON"``,
              * ``"MULTIPOINT"``,
              * ``"MULTILINESTRING"``,
              * ``"MULTIPOLYGON"``,
              * ``"GEOMETRYCOLLECTION"``,
              * ``"CURVE"``,
              * ``None``.

           The latter is actually not supported with
           :class:`geoalchemy2.types.Geography`.

           When set to ``None`` then no "geometry type" constraints will be
           attached to the geometry type declaration. Using ``None`` here
           is not compatible with setting ``management`` to ``True``.

           Default is ``"GEOMETRY"``.

        srid: The SRID for this column. E.g. 4326. Default is ``-1``.
        dimension: The dimension of the geometry. Default is ``2``.

            With ``management`` set to ``True``, that is when ``AddGeometryColumn`` is used
            to add the geometry column, there are two constraints:

            * The ``geometry_type`` must not end with ``"ZM"``.  This is due to PostGIS'
              ``AddGeometryColumn`` failing with ZM geometry types. Instead the "simple"
              geometry type (e.g. POINT rather POINTZM) should be used with ``dimension``
              set to ``4``.
            * When the ``geometry_type`` ends with ``"Z"`` or ``"M"`` then ``dimension``
              must be set to ``3``.

            With ``management`` set to ``False`` (the default) ``dimension`` is not
            taken into account, and the actual dimension is fully defined with the
            ``geometry_type``.

        spatial_index: Indicate if a spatial index should be created. Default is ``True``.
        use_N_D_index: Use the N-D index instead of the standard 2-D index.
        management: Indicate if the ``AddGeometryColumn`` and ``DropGeometryColumn``
            managements functions should be called when adding and dropping the
            geometry column. Should be set to ``True`` for PostGIS 1.x and SQLite. Default is
            ``False``. Note that this option has no effect for
            :class:`geoalchemy2.types.Geography`.
        use_typmod: By default PostgreSQL type modifiers are used to create the geometry
            column. To use check constraints instead set ``use_typmod`` to
            ``False``. By default this option is not included in the call to
            ``AddGeometryColumn``. Note that this option is only taken
            into account if ``management`` is set to ``True`` and is only available
            for PostGIS 2.x.
    """

    name = None
    """ Name used for defining the main geo type (geometry or geography)
        in CREATE TABLE statements. Set in subclasses. """

    from_text = None
    """ The name of "from text" function for this type.
        Set in subclasses. """

    as_binary = None
    """ The name of the "as binary" function for this type.
        Set in subclasses. """

    comparator_factory = Comparator
    """ This is the way by which spatial operators are defined for
        geometry/geography columns. """

    cache_ok = False
    """ Disable cache for this type. """

    def __init__(self, geometry_type='GEOMETRY', srid=-1, dimension=2,
                 spatial_index=True, use_N_D_index=False, management=False, use_typmod=None,
                 from_text=None, name=None, nullable=True, _spatial_index_reflected=None):
        geometry_type, srid = self.check_ctor_args(
            geometry_type, srid, dimension, management, use_typmod, nullable)
        self.geometry_type = geometry_type
        self.srid = srid
        if name is not None:
            self.name = name
        if from_text is not None:
            self.from_text = from_text
        self.dimension = dimension
        self.spatial_index = spatial_index
        self.use_N_D_index = use_N_D_index
        self.management = management
        self.use_typmod = use_typmod
        self.extended = self.as_binary == 'ST_AsEWKB'
        self.nullable = nullable
        self._spatial_index_reflected = _spatial_index_reflected

    def get_col_spec(self):
        if not self.geometry_type:
            return self.name
        return '%s(%s,%d)' % (self.name, self.geometry_type, self.srid)

    def column_expression(self, col):
        """Specific column_expression that automatically adds a conversion function"""
        return getattr(func, self.as_binary)(col, type_=self)

    def result_processor(self, dialect, coltype):
        """Specific result_processor that automatically process spatial elements"""
        def process(value):
            if value is not None:
                kwargs = {}
                if self.srid > 0:
                    kwargs['srid'] = self.srid
                if self.extended is not None:
                    kwargs['extended'] = self.extended
                return self.ElementType(value, **kwargs)
        return process

    def bind_expression(self, bindvalue):
        """Specific bind_expression that automatically adds a conversion function"""
        return getattr(func, self.from_text)(bindvalue, type_=self)

    def bind_processor(self, dialect):
        """Specific bind_processor that automatically process spatial elements"""
        def process(bindvalue):
            if isinstance(bindvalue, WKTElement):
                if bindvalue.extended:
                    return '%s' % (bindvalue.data)
                else:
                    return 'SRID=%d;%s' % (bindvalue.srid, bindvalue.data)
            elif isinstance(bindvalue, WKBElement):
                if dialect.name == 'sqlite' or not bindvalue.extended:
                    # With SpatiaLite or when the WKBElement includes a WKB value rather
                    # than a EWKB value we use Shapely to convert the WKBElement to an
                    # EWKT string
                    if not SHAPELY:
                        raise ArgumentError('Shapely is required for handling WKBElement bind '
                                            'values when using SpatiaLite or when the bind value '
                                            'is a WKB rather than an EWKB')
                    shape = to_shape(bindvalue)
                    return 'SRID=%d;%s' % (bindvalue.srid, shape.wkt)
                else:
                    # PostGIS ST_GeomFromEWKT works with EWKT strings as well
                    # as EWKB hex strings
                    return bindvalue.desc
            elif isinstance(bindvalue, RasterElement):
                return '%s' % (bindvalue.data)
            else:
                return bindvalue
        return process

    @staticmethod
    def check_ctor_args(geometry_type, srid, dimension, management, use_typmod, nullable):
        try:
            srid = int(srid)
        except ValueError:
            raise ArgumentError('srid must be convertible to an integer')
        if geometry_type:
            geometry_type = geometry_type.upper()
            if management:
                if geometry_type.endswith('ZM'):
                    # PostGIS' AddGeometryColumn does not work with ZM geometry types. Instead
                    # the simple geometry type (e.g. POINT rather POINTZM) should be used with
                    # dimension set to 4
                    raise ArgumentError(
                        'with management=True use geometry_type={!r} and '
                        'dimension=4 for {!r} geometries'.format(geometry_type[:-2], geometry_type))
                elif geometry_type[-1] in ('Z', 'M') and dimension != 3:
                    # If a Z or M geometry type is used then dimension must be set to 3
                    raise ArgumentError(
                        'with management=True dimension must be 3 for '
                        '{!r} geometries'.format(geometry_type))
        else:
            if management:
                raise ArgumentError('geometry_type set to None not compatible '
                                    'with management')
            if srid > 0:
                warnings.warn('srid not enforced when geometry_type is None')

        if use_typmod and not management:
            warnings.warn('use_typmod ignored when management is False')
        if use_typmod is not None and not nullable:
            raise ArgumentError(
                'The "nullable" and "use_typmod" arguments can not be used together'
            )

        return geometry_type, srid


class Geometry(_GISType):
    """
    The Geometry type.

    Creating a geometry column is done like this::

        Column(Geometry(geometry_type='POINT', srid=4326))

    See :class:`geoalchemy2.types._GISType` for the list of arguments that can
    be passed to the constructor.

    If ``srid`` is set then the ``WKBElement`` objects resulting from queries will
    have that SRID, and, when constructing the ``WKBElement`` objects, the SRID
    won't be read from the data returned by the database. If ``srid`` is not set
    (meaning it's ``-1``) then the SRID set in ``WKBElement`` objects will be read
    from the data returned by the database.
    """

    name = 'geometry'
    """ Type name used for defining geometry columns in ``CREATE TABLE``. """

    from_text = 'ST_GeomFromEWKT'
    """ The "from text" geometry constructor. Used by the parent class'
        ``bind_expression`` method. """

    as_binary = 'ST_AsEWKB'
    """ The "as binary" function to use. Used by the parent class'
        ``column_expression`` method. """

    ElementType = WKBElement
    """ The element class to use. Used by the parent class'
        ``result_processor`` method. """

    cache_ok = False
    """ Disable cache for this type. """


class Geography(_GISType):
    """
    The Geography type.

    Creating a geography column is done like this::

        Column(Geography(geometry_type='POINT', srid=4326))

    See :class:`geoalchemy2.types._GISType` for the list of arguments that can
    be passed to the constructor.

    """

    name = 'geography'
    """ Type name used for defining geography columns in ``CREATE TABLE``. """

    from_text = 'ST_GeogFromText'
    """ The ``FromText`` geography constructor. Used by the parent class'
        ``bind_expression`` method. """

    as_binary = 'ST_AsBinary'
    """ The "as binary" function to use. Used by the parent class'
        ``column_expression`` method. """

    ElementType = WKBElement
    """ The element class to use. Used by the parent class'
        ``result_processor`` method. """

    cache_ok = False
    """ Disable cache for this type. """


class Raster(_GISType):
    """
    The Raster column type.

    Creating a raster column is done like this::

        Column(Raster)

    This class defines the ``result_processor`` method, so that raster values
    received from the database are converted to
    :class:`geoalchemy2.elements.RasterElement` objects.

    Args:
        spatial_index: Indicate if a spatial index should be created. Default is ``True``.

    """

    comparator_factory = BaseComparator
    """
    This is the way by which spatial operators and functions are
    defined for raster columns.
    """

    name = 'raster'
    """ Type name used for defining raster columns in ``CREATE TABLE``. """

    from_text = 'raster'
    """ The "from text" raster constructor. Used by the parent class'
        ``bind_expression`` method. """

    as_binary = 'raster'
    """ The "as binary" function to use. Used by the parent class'
        ``column_expression`` method. """

    ElementType = RasterElement
    """ The element class to use. Used by the parent class'
        ``result_processor`` method. """

    cache_ok = False
    """ Disable cache for this type. """

    def __init__(self, spatial_index=True, from_text=None, name=None, nullable=True):
        # Enforce default values
        super(Raster, self).__init__(
            geometry_type=None,
            srid=-1,
            dimension=2,
            spatial_index=spatial_index,
            use_N_D_index=False,
            management=False,
            use_typmod=False,
            from_text=from_text,
            name=name,
            nullable=nullable,
        )
        self.extended = None


class _DummyGeometry(Geometry):
    """A dummy type only used with SQLite."""

    def get_col_spec(self):
        return 'GEOMETRY'


class CompositeType(UserDefinedType):
    """
    A wrapper for :class:`geoalchemy2.elements.CompositeElement`, that can be
    used as the return type in PostgreSQL functions that return composite
    values.

    This is used as the base class of :class:`geoalchemy2.types.GeometryDump`.
    """

    typemap = {}
    """ Dictionary used for defining the content types and their
        corresponding keys. Set in subclasses. """

    class comparator_factory(UserDefinedType.Comparator):
        def __getattr__(self, key):
            try:
                type_ = self.type.typemap[key]
            except KeyError:
                raise AttributeError("Type '%s' doesn't have an attribute: '%s'"
                                     % (self.type, key))

            return CompositeElement(self.expr, key, type_)


class GeometryDump(CompositeType):
    """
    The return type for functions like ``ST_Dump``, consisting of a path and
    a geom field. You should normally never use this class directly.
    """

    typemap = {'path': postgresql.ARRAY(Integer), 'geom': Geometry}
    """ Dictionary defining the contents of a ``geometry_dump``. """

    cache_ok = True
    """ Enable cache for this type. """


# Register Geometry, Geography and Raster to SQLAlchemy's reflection subsystems.
postgresql_ischema_names['geometry'] = Geometry
postgresql_ischema_names['geography'] = Geography
postgresql_ischema_names['raster'] = Raster

sqlite_ischema_names['GEOMETRY'] = Geometry
sqlite_ischema_names['POINT'] = Geometry
sqlite_ischema_names['LINESTRING'] = Geometry
sqlite_ischema_names['POLYGON'] = Geometry
sqlite_ischema_names['MULTIPOINT'] = Geometry
sqlite_ischema_names['MULTILINESTRING'] = Geometry
sqlite_ischema_names['MULTIPOLYGON'] = Geometry
sqlite_ischema_names['CURVE'] = Geometry
sqlite_ischema_names['GEOMETRYCOLLECTION'] = Geometry
sqlite_ischema_names['RASTER'] = Raster


class SummaryStats(CompositeType):
    """Define the composite type returned by the function ST_SummaryStatsAgg"""
    typemap = {
        'count': Integer,
        'sum': Float,
        'mean': Float,
        'stddev': Float,
        'min': Float,
        'max': Float,
    }

    cache_ok = True
    """ Enable cache for this type. """



"""
Disable wrapping in select
==========================

If the application wants to build queries with GeoAlchemy 2 and gets them as strings,
the wrapping of geometry columns with a `ST_AsEWKB()` function might be annoying. In
this case it is possible to disable this wrapping.
This example uses SQLAlchemy ORM queries.
"""
from sqlalchemy import Column
from sqlalchemy import Integer
from sqlalchemy import func
from sqlalchemy.ext.declarative import declarative_base

from geoalchemy2 import Geometry

# Tests imports
from tests import select

Base = declarative_base()


class RawGeometry(Geometry):
    """This class is used to remove the 'ST_AsEWKB()'' function from select queries"""

    def column_expression(self, col):
        return col


class Point(Base):
    __tablename__ = "point"
    id = Column(Integer, primary_key=True)
    geom = Column(Geometry(srid=4326, geometry_type="POINT"))
    raw_geom = Column(
        RawGeometry(srid=4326, geometry_type="POINT"))


def test_no_wrapping():
    # Select all columns
    select_query = select([Point])

    # Check that the 'geom' column is wrapped by 'ST_AsEWKB()' and that the column
    # 'raw_geom' is not.
    assert str(select_query) == (
        "SELECT point.id, ST_AsEWKB(point.geom) AS geom, point.raw_geom \n"
        "FROM point"
    )


def test_func_no_wrapping():
    # Select query with function
    select_query = select([
        func.ST_Buffer(Point.geom),  # with wrapping (default behavior)
        func.ST_Buffer(Point.geom, type_=Geometry),  # with wrapping
        func.ST_Buffer(Point.geom, type_=RawGeometry)  # without wrapping
    ])

    # Check the query
    assert str(select_query) == (
        "SELECT "
        "ST_AsEWKB(ST_Buffer(point.geom)) AS \"ST_Buffer_1\", "
        "ST_AsEWKB(ST_Buffer(point.geom)) AS \"ST_Buffer_2\", "
        "ST_Buffer(point.geom) AS \"ST_Buffer_3\" \n"
        "FROM point"
    )



"""
Reproject a Raster using ST_Transform
=====================================

The `ST_Transform()` function (and a few others like `ST_SnapToGrid()`) can be used on
both `Geometry` and `Raster` types. In `GeoAlchemy2`, this function is only defined for
`Geometry` as it can not be defined for several types at the same time. Thus using this
function on `Raster` requires minor tweaking.

This example uses both SQLAlchemy core and ORM queries.
"""
from sqlalchemy import Column
from sqlalchemy import Integer
from sqlalchemy import MetaData
from sqlalchemy import Table
from sqlalchemy import func
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import Query

from geoalchemy2 import Geometry
from geoalchemy2 import Raster

# Tests imports
from tests import select

metadata = MetaData()
Base = declarative_base(metadata=metadata)

table = Table(
    "raster_table",
    metadata,
    Column("id", Integer, primary_key=True),
    Column("geom", Geometry("POLYGON", 4326)),
    Column("rast", Raster()),
)


class RasterTable(Base):
    __tablename__ = 'raster_table_orm'
    id = Column(Integer, primary_key=True)
    geom = Column(Geometry("POLYGON", 4326))
    rast = Column(Raster())

    def __init__(self, rast):
        self.rast = rast


def test_transform_core():
    # Define the transform query for both the geometry and the raster in a naive way
    wrong_query = select([
        func.ST_Transform(table.c.geom, 2154),
        func.ST_Transform(table.c.rast, 2154)
    ])

    # Check the query
    assert str(wrong_query) == (
        "SELECT "
        "ST_AsEWKB("
        "ST_Transform(raster_table.geom, :ST_Transform_2)) AS \"ST_Transform_1\", "
        "ST_AsEWKB("  # <= Note that the raster is processed as a Geometry here
        "ST_Transform(raster_table.rast, :ST_Transform_4)) AS \"ST_Transform_3\" \n"
        "FROM raster_table"
    )

    # Define the transform query for both the geometry and the raster in the correct way
    correct_query = select([
        func.ST_Transform(table.c.geom, 2154),
        func.ST_Transform(table.c.rast, 2154, type_=Raster)
    ])

    # Check the query
    assert str(correct_query) == (
        "SELECT "
        "ST_AsEWKB("
        "ST_Transform(raster_table.geom, :ST_Transform_2)) AS \"ST_Transform_1\", "
        "raster("  # <= This time the raster is correctly processed as a Raster
        "ST_Transform(raster_table.rast, :ST_Transform_4)) AS \"ST_Transform_3\" \n"
        "FROM raster_table"
    )


def test_transform_ORM():
    # Define the transform query for both the geometry and the raster in a naive way
    wrong_query = Query([
        RasterTable.geom.ST_Transform(2154),
        RasterTable.rast.ST_Transform(2154)
    ])

    # Check the query
    assert str(wrong_query) == (
        "SELECT "
        "ST_AsEWKB("
        "ST_Transform(raster_table_orm.geom, :ST_Transform_2)) AS \"ST_Transform_1\", "
        "ST_AsEWKB("  # <= Note that the raster is processed as a Geometry here
        "ST_Transform(raster_table_orm.rast, :ST_Transform_4)) AS \"ST_Transform_3\" \n"
        "FROM raster_table_orm"
    )

    # Define the transform query for both the geometry and the raster in the correct way
    correct_query = Query([
        RasterTable.geom.ST_Transform(2154),
        RasterTable.rast.ST_Transform(2154, type_=Raster)
    ])

    # Check the query
    assert str(correct_query) == (
        "SELECT "
        "ST_AsEWKB("
        "ST_Transform(raster_table_orm.geom, :ST_Transform_2)) AS \"ST_Transform_1\", "
        "raster("  # <= This time the raster is correctly processed as a Raster
        "ST_Transform(raster_table_orm.rast, :ST_Transform_4)) AS \"ST_Transform_3\" \n"
        "FROM raster_table_orm"
    )



"""
Use CompositeType
=================

Some functions return composite types. This example shows how to deal with this
kind of functions.
"""
import pytest
from pkg_resources import parse_version
from sqlalchemy import Column
from sqlalchemy import Float
from sqlalchemy import Integer
from sqlalchemy import MetaData
from sqlalchemy import __version__ as SA_VERSION
from sqlalchemy.ext.declarative import declarative_base

from geoalchemy2 import Raster
from geoalchemy2 import WKTElement
from geoalchemy2.functions import GenericFunction
from geoalchemy2.types import CompositeType

# Tests imports
from tests import select
from tests import test_only_with_dialects


class SummaryStatsCustomType(CompositeType):
    """Define the composite type returned by the function ST_SummaryStatsAgg."""
    typemap = {
        'count': Integer,
        'sum': Float,
        'mean': Float,
        'stddev': Float,
        'min': Float,
        'max': Float,
    }

    cache_ok = True


class ST_SummaryStatsAgg(GenericFunction):
    type = SummaryStatsCustomType
    # Set a specific identifier to not override the actual ST_SummaryStatsAgg function
    identifier = "ST_SummaryStatsAgg_custom"

    inherit_cache = True


metadata = MetaData()
Base = declarative_base(metadata=metadata)


class Ocean(Base):
    __tablename__ = 'ocean'
    id = Column(Integer, primary_key=True)
    rast = Column(Raster)

    def __init__(self, rast):
        self.rast = rast


@test_only_with_dialects("postgresql")
class TestSTSummaryStatsAgg():

    @pytest.mark.skipif(
        parse_version(SA_VERSION) < parse_version("1.4"),
        reason="requires SQLAlchely>1.4",
    )
    def test_st_summary_stats_agg(self, session, conn):
        metadata.drop_all(conn, checkfirst=True)
        metadata.create_all(conn)

        # Create a new raster
        polygon = WKTElement('POLYGON((0 0,1 1,0 1,0 0))', srid=4326)
        o = Ocean(polygon.ST_AsRaster(5, 6))
        session.add(o)
        session.flush()

        # Define the query to compute stats
        stats_agg = select([
            Ocean.rast.ST_SummaryStatsAgg_custom(1, True, 1).label("stats")
        ])
        stats_agg_alias = stats_agg.alias("stats_agg")

        # Use these stats
        query = select([
            stats_agg_alias.c.stats.count.label("count"),
            stats_agg_alias.c.stats.sum.label("sum"),
            stats_agg_alias.c.stats.mean.label("mean"),
            stats_agg_alias.c.stats.stddev.label("stddev"),
            stats_agg_alias.c.stats.min.label("min"),
            stats_agg_alias.c.stats.max.label("max")
        ])

        # Check the query
        assert str(query.compile(dialect=session.bind.dialect)) == (
            "SELECT "
            "(stats_agg.stats).count AS count, "
            "(stats_agg.stats).sum AS sum, "
            "(stats_agg.stats).mean AS mean, "
            "(stats_agg.stats).stddev AS stddev, "
            "(stats_agg.stats).min AS min, "
            "(stats_agg.stats).max AS max \n"
            "FROM ("
            "SELECT "
            "ST_SummaryStatsAgg("
            "ocean.rast, "
            "%(ST_SummaryStatsAgg_1)s, %(ST_SummaryStatsAgg_2)s, %(ST_SummaryStatsAgg_3)s"
            ") AS stats \n"
            "FROM ocean) AS stats_agg"
        )

        # Execute the query
        res = session.execute(query).fetchall()

        # Check the result
        assert res == [(15, 15.0, 1.0, 0.0, 1.0, 1.0)]



# -*- coding: utf-8 -*-
# flake8: noqa
from . import types

_FUNCTIONS = [
    ('AddGeometryColumn', None,
     '''Adds a geometry column to an existing table.'''),
    ('DropGeometryColumn', None,
     '''Removes a geometry column from a spatial table.'''),
    ('DropGeometryTable', None,
     '''Drops a table and all its references in geometry_columns.'''),
    ('Find_SRID', None,
     '''Returns the SRID defined for a geometry column.'''),
    ('Populate_Geometry_Columns', None,
     '''Ensures geometry columns are defined with type modifiers or have appropriate spatial constraints.'''),
    ('UpdateGeometrySRID', None,
     '''Updates the SRID of all features in a geometry column, and the table metadata.'''),
    ('ST_Collect', types.Geometry,
     '''Creates a GeometryCollection or Multi* geometry from a set of geometries.'''),
    ('ST_LineFromMultiPoint', types.Geometry,
     '''Creates a LineString from a MultiPoint geometry.'''),
    ('ST_MakeEnvelope', types.Geometry,
     '''Creates a rectangular Polygon from minimum and maximum coordinates.'''),
    ('ST_MakeLine', types.Geometry,
     '''Creates a Linestring from Point, MultiPoint, or LineString geometries.'''),
    ('ST_MakePoint', types.Geometry,
     '''Creates a 2D, 3DZ or 4D Point.'''),
    ('ST_MakePointM', types.Geometry,
     '''Creates a Point from X, Y and M values.'''),
    ('ST_MakePolygon', types.Geometry,
     '''Creates a Polygon from a shell and optional list of holes.'''),
    ('ST_Point', types.Geometry,
     '''Creates a Point with the given coordinate values. Alias for ST_MakePoint.'''),
    ('ST_Polygon', types.Geometry,
     '''[geometry] Creates a Polygon from a LineString with a specified SRID.\nOR\n[raster] Returns a multipolygon geometry formed by the union of pixels that have a pixel value that is not no data value. If no band number is specified, band num defaults to 1.'''),
    ('ST_TileEnvelope', types.Geometry,
     '''Creates a rectangular Polygon in Web Mercator (SRID:3857) using the XYZ tile system.'''),
    ('GeometryType', None,
     '''Returns the type of a geometry as text.'''),
    ('ST_Boundary', types.Geometry,
     '''Returns the boundary of a geometry.'''),
    ('ST_CoordDim', None,
     '''Return the coordinate dimension of a geometry.'''),
    ('ST_Dimension', None,
     '''Returns the topological dimension of a geometry.'''),
    ('ST_Dump', types.GeometryDump,
     '''Returns a set of geometry_dump rows for the components of a geometry.'''),
    ('ST_DumpPoints', types.GeometryDump,
     '''Returns a set of geometry_dump rows for the points in a geometry.'''),
    ('ST_DumpRings', types.GeometryDump,
     '''Returns a set of geometry_dump rows for the exterior and interior rings of a Polygon.'''),
    ('ST_EndPoint', types.Geometry,
     '''Returns the last point of a LineString or CircularLineString.'''),
    ('ST_Envelope', types.Geometry,
     '''[geometry] Returns a geometry representing the bounding box of a geometry.\nOR\n[raster] Returns the polygon representation of the extent of the raster.'''),
    ('ST_BoundingDiagonal', types.Geometry,
     '''Returns the diagonal of a geometry's bounding box.'''),
    ('ST_ExteriorRing', types.Geometry,
     '''Returns a LineString representing the exterior ring of a Polygon.'''),
    ('ST_GeometryN', types.Geometry,
     '''Return the Nth geometry element of a geometry collection.'''),
    ('ST_GeometryType', None,
     '''Returns the SQL-MM type of a geometry as text.'''),
    ('ST_HasArc', None,
     '''Tests if a geometry contains a circular arc'''),
    ('ST_InteriorRingN', types.Geometry,
     '''Returns the Nth interior ring (hole) of a Polygon.'''),
    ('ST_IsPolygonCCW', None,
     '''Tests if Polygons have exterior rings oriented counter-clockwise and interior rings oriented clockwise.'''),
    ('ST_IsPolygonCW', None,
     '''Tests if Polygons have exterior rings oriented clockwise and interior rings oriented counter-clockwise.'''),
    ('ST_IsClosed', None,
     '''Tests if a LineStrings's start and end points are coincident. For a PolyhedralSurface tests if it is closed (volumetric).'''),
    ('ST_IsCollection', None,
     '''Tests if a geometry is a geometry collection type.'''),
    ('ST_IsEmpty', None,
     '''[geometry] Tests if a geometry is empty.\nOR\n[raster] Returns true if the raster is empty (width = 0 and height = 0). Otherwise, returns false.'''),
    ('ST_IsRing', None,
     '''Tests if a LineString is closed and simple.'''),
    ('ST_IsSimple', None,
     '''Tests if a geometry has no points of self-intersection or self-tangency.'''),
    ('ST_M', None,
     '''Returns the M coordinate of a Point.'''),
    ('ST_MemSize', None,
     '''[geometry] Returns the amount of memory space a geometry takes.\nOR\n[raster] Returns the amount of space (in bytes) the raster takes.'''),
    ('ST_NDims', None,
     '''Returns the coordinate dimension of a geometry.'''),
    ('ST_NPoints', None,
     '''Returns the number of points (vertices) in a geometry.'''),
    ('ST_NRings', None,
     '''Returns the number of rings in a polygonal geometry.'''),
    ('ST_NumGeometries', None,
     '''Returns the number of elements in a geometry collection.'''),
    ('ST_NumInteriorRings', None,
     '''Returns the number of interior rings (holes) of a Polygon.'''),
    ('ST_NumInteriorRing', None,
     '''Returns the number of interior rings (holes) of a Polygon. Aias for ST_NumInteriorRings'''),
    ('ST_NumPatches', None,
     '''Return the number of faces on a Polyhedral Surface. Will return null for non-polyhedral geometries.'''),
    ('ST_NumPoints', None,
     '''Returns the number of points in a LineString or CircularString.'''),
    ('ST_PatchN', types.Geometry,
     '''Returns the Nth geometry (face) of a PolyhedralSurface.'''),
    ('ST_PointN', types.Geometry,
     '''Returns the Nth point in the first LineString or circular LineString in a geometry.'''),
    ('ST_Points', types.Geometry,
     '''Returns a MultiPoint containing all the coordinates of a geometry.'''),
    ('ST_StartPoint', types.Geometry,
     '''Returns the first point of a LineString.'''),
    ('ST_Summary', None,
     '''[geometry] Returns a text summary of the contents of a geometry.\nOR\n[raster] Returns a text summary of the contents of the raster.'''),
    ('ST_X', None,
     '''Returns the X coordinate of a Point.'''),
    ('ST_Y', None,
     '''Returns the Y coordinate of a Point.'''),
    ('ST_Z', None,
     '''Returns the Z coordinate of a Point.'''),
    ('ST_Zmflag', None,
     '''Returns a code indicating the ZM coordinate dimension of a geometry.'''),
    ('ST_AddPoint', types.Geometry,
     '''Add a point to a LineString.'''),
    ('ST_CollectionExtract', types.Geometry,
     '''Given a (multi)geometry, return a (multi)geometry consisting only of elements of the specified type.'''),
    ('ST_CollectionHomogenize', types.Geometry,
     '''Given a geometry collection, return the \"simplest\" representation of the contents.'''),
    ('ST_Force2D', types.Geometry,
     '''Force the geometries into a \"2-dimensional mode\".'''),
    ('ST_Force3D', types.Geometry,
     ('''Force the geometries into XYZ mode. This is an alias for ST_Force3DZ.''', 'ST_Force_3D')),
    ('ST_Force3DZ', types.Geometry,
     ('''Force the geometries into XYZ mode.''', 'ST_Force_3DZ')),
    ('ST_Force3DM', types.Geometry,
     ('''Force the geometries into XYM mode.''', 'ST_Force_3DZ')),
    ('ST_Force4D', types.Geometry,
     ('''Force the geometries into XYZM mode.''', 'ST_Force_4D')),
    ('ST_ForcePolygonCCW', types.Geometry,
     '''Orients all exterior rings counter-clockwise and all interior rings clockwise.'''),
    ('ST_ForceCollection', types.Geometry,
     ('''Convert the geometry into a GEOMETRYCOLLECTION.''', 'ST_Force_Collection')),
    ('ST_ForcePolygonCW', types.Geometry,
     '''Orients all exterior rings clockwise and all interior rings counter-clockwise.'''),
    ('ST_ForceSFS', types.Geometry,
     '''Force the geometries to use SFS 1.1 geometry types only.'''),
    ('ST_ForceRHR', types.Geometry,
     '''Force the orientation of the vertices in a polygon to follow the Right-Hand-Rule.'''),
    ('ST_ForceCurve', types.Geometry,
     '''Upcast a geometry into its curved type, if applicable.'''),
    ('ST_LineMerge', types.Geometry,
     '''Return a (set of) LineString(s) formed by sewing together a MULTILINESTRING.'''),
    ('ST_Multi', types.Geometry,
     '''Return the geometry as a MULTI* geometry.'''),
    ('ST_Normalize', types.Geometry,
     '''Return the geometry in its canonical form.'''),
    ('ST_QuantizeCoordinates', types.Geometry,
     '''Sets least significant bits of coordinates to zero'''),
    ('ST_RemovePoint', types.Geometry,
     '''Remove point from a linestring.'''),
    ('ST_Reverse', types.Geometry,
     '''Return the geometry with vertex order reversed.'''),
    ('ST_Segmentize', types.Geometry,
     '''Return a modified geometry/geography having no segment longer than the given distance.'''),
    ('ST_SetPoint', types.Geometry,
     '''Replace point of a linestring with a given point.'''),
    ('ST_SnapToGrid', types.Geometry,
     '''[geometry] Snap all points of the input geometry to a regular grid.\nOR\n[raster] Resample a raster by snapping it to a grid. New pixel values are computed using the NearestNeighbor (english or american spelling), Bilinear, Cubic, CubicSpline or Lanczos resampling algorithm. Default is NearestNeighbor.'''),
    ('ST_Snap', types.Geometry,
     '''Snap segments and vertices of input geometry to vertices of a reference geometry.'''),
    ('ST_SwapOrdinates', types.Geometry,
     '''Returns a version of the given geometry with given ordinate values swapped.'''),
    ('ST_IsValid', None,
     '''Tests if a geometry is well-formed in 2D.'''),
    ('ST_IsValidDetail', None,
     '''Returns a valid_detail row stating if a geometry is valid, and if not a reason why and a location.'''),
    ('ST_IsValidReason', None,
     '''Returns text stating if a geometry is valid, or a reason for invalidity.'''),
    ('ST_SetSRID', types.Geometry,
     '''[geometry] Set the SRID on a geometry to a particular integer value.\nOR\n[raster] Sets the SRID of a raster to a particular integer srid defined in the spatial_ref_sys table.'''),
    ('ST_SRID', None,
     '''[geometry] Returns the spatial reference identifier for the ST_Geometry as defined in spatial_ref_sys table.\nOR\n[raster] Returns the spatial reference identifier of the raster as defined in spatial_ref_sys table.'''),
    ('ST_Transform', types.Geometry,
     '''[geometry] Return a new geometry with its coordinates transformed to a different spatial reference system.\nOR\n[raster] Reprojects a raster in a known spatial reference system to another known spatial reference system using specified resampling algorithm. Options are NearestNeighbor, Bilinear, Cubic, CubicSpline, Lanczos defaulting to NearestNeighbor.'''),
    ('ST_BdPolyFromText', types.Geometry,
     '''Construct a Polygon given an arbitrary collection of closed linestrings as a MultiLineString Well-Known text representation.'''),
    ('ST_BdMPolyFromText', types.Geometry,
     '''Construct a MultiPolygon given an arbitrary collection of closed linestrings as a MultiLineString text representation Well-Known text representation.'''),
    ('ST_GeogFromText', types.Geography,
     '''Return a specified geography value from Well-Known Text representation or extended (WKT).'''),
    ('ST_GeographyFromText', types.Geography,
     '''Return a specified geography value from Well-Known Text representation or extended (WKT).'''),
    ('ST_GeomCollFromText', types.Geometry,
     '''Makes a collection Geometry from collection WKT with the given SRID. If SRID is not given, it defaults to 0.'''),
    ('ST_GeomFromEWKT', types.Geometry,
     '''Return a specified ST_Geometry value from Extended Well-Known Text representation (EWKT).'''),
    ('ST_GeometryFromText', types.Geometry,
     '''Return a specified ST_Geometry value from Well-Known Text representation (WKT). This is an alias name for ST_GeomFromText'''),
    ('ST_GeomFromText', types.Geometry,
     '''Return a specified ST_Geometry value from Well-Known Text representation (WKT).'''),
    ('ST_LineFromText', types.Geometry,
     '''Makes a Geometry from WKT representation with the given SRID. If SRID is not given, it defaults to 0.'''),
    ('ST_MLineFromText', types.Geometry,
     '''Return a specified ST_MultiLineString value from WKT representation.'''),
    ('ST_MPointFromText', types.Geometry,
     '''Makes a Geometry from WKT with the given SRID. If SRID is not given, it defaults to 0.'''),
    ('ST_MPolyFromText', types.Geometry,
     '''Makes a MultiPolygon Geometry from WKT with the given SRID. If SRID is not given, it defaults to 0.'''),
    ('ST_PointFromText', types.Geometry,
     '''Makes a point Geometry from WKT with the given SRID. If SRID is not given, it defaults to unknown.'''),
    ('ST_PolygonFromText', types.Geometry,
     '''Makes a Geometry from WKT with the given SRID. If SRID is not given, it defaults to 0.'''),
    ('ST_WKTToSQL', types.Geometry,
     '''Return a specified ST_Geometry value from Well-Known Text representation (WKT). This is an alias name for ST_GeomFromText'''),
    ('ST_GeogFromWKB', types.Geography,
     '''Creates a geography instance from a Well-Known Binary geometry representation (WKB) or extended Well Known Binary (EWKB).'''),
    ('ST_GeomFromEWKB', types.Geometry,
     '''Return a specified ST_Geometry value from Extended Well-Known Binary representation (EWKB).'''),
    ('ST_GeomFromWKB', types.Geometry,
     '''Creates a geometry instance from a Well-Known Binary geometry representation (WKB) and optional SRID.'''),
    ('ST_LineFromWKB', types.Geometry,
     '''Makes a LINESTRING from WKB with the given SRID'''),
    ('ST_LinestringFromWKB', types.Geometry,
     '''Makes a geometry from WKB with the given SRID.'''),
    ('ST_PointFromWKB', types.Geometry,
     '''Makes a geometry from WKB with the given SRID'''),
    ('ST_WKBToSQL', types.Geometry,
     '''Return a specified ST_Geometry value from Well-Known Binary representation (WKB). This is an alias name for ST_GeomFromWKB that takes no srid'''),
    ('ST_Box2dFromGeoHash', None,  # return an unsupported Box2d object
     '''Return a BOX2D from a GeoHash string.'''),
    ('ST_GeomFromGeoHash', types.Geometry,
     '''Return a geometry from a GeoHash string.'''),
    ('ST_GeomFromGML', types.Geometry,
     '''Takes as input GML representation of geometry and outputs a PostGIS geometry object'''),
    ('ST_GeomFromGeoJSON', types.Geometry,
     '''Takes as input a geojson representation of a geometry and outputs a PostGIS geometry object'''),
    ('ST_GeomFromKML', types.Geometry,
     '''Takes as input KML representation of geometry and outputs a PostGIS geometry object'''),
    ('ST_GeomFromTWKB', types.Geometry,
     '''Creates a geometry instance from a TWKB (\"Tiny Well-Known Binary\") geometry representation.'''),
    ('ST_GMLToSQL', types.Geometry,
     '''Return a specified ST_Geometry value from GML representation. This is an alias name for ST_GeomFromGML'''),
    ('ST_LineFromEncodedPolyline', types.Geometry,
     '''Creates a LineString from an Encoded Polyline.'''),
    ('ST_PointFromGeoHash', types.Geometry,
     '''Return a point from a GeoHash string.'''),
    ('ST_AsEWKT', None,
     '''Return the Well-Known Text (WKT) representation of the geometry with SRID meta data.'''),
    ('ST_AsText', None,
     '''Return the Well-Known Text (WKT) representation of the geometry/geography without SRID metadata.'''),
    ('ST_AsBinary', None,
     '''[gometry] Return the Well-Known Binary (WKB) representation of the geometry/geography without SRID meta data.\nOR\n[raster] Return the Well-Known Binary (WKB) representation of the raster.'''),
    ('ST_AsEWKB', None,
     '''Return the Well-Known Binary (WKB) representation of the geometry with SRID meta data.'''),
    ('ST_AsHEXEWKB', None,
     '''Returns a Geometry in HEXEWKB format (as text) using either little-endian (NDR) or big-endian (XDR) encoding.'''),
    ('ST_AsEncodedPolyline', None,
     '''Returns an Encoded Polyline from a LineString geometry.'''),
    ('ST_AsGeobuf', None,
     '''Return a Geobuf representation of a set of rows.'''),
    ('ST_AsGML', None,
     '''Return the geometry as a GML version 2 or 3 element.'''),
    ('ST_AsKML', None,
     '''Return the geometry as a KML element. Several variants. Default version=2, default maxdecimaldigits=15'''),
    ('ST_AsLatLonText', None,
     '''Return the Degrees, Minutes, Seconds representation of the given point.'''),
    ('ST_AsMVTGeom', types.Geometry,
     '''Transform a geometry into the coordinate space of a Mapbox Vector Tile.'''),
    ('ST_AsMVT', None,
     '''Aggregate function returning a Mapbox Vector Tile representation of a set of rows.'''),
    ('ST_AsSVG', None,
     '''Returns SVG path data for a geometry.'''),
    ('ST_AsTWKB', None,
     '''Returns the geometry as TWKB, aka \"Tiny Well-Known Binary\"'''),
    ('ST_AsX3D', None,
     '''Returns a Geometry in X3D xml node element format: ISO-IEC-19776-1.2-X3DEncodings-XML'''),
    ('ST_GeoHash', None,
     '''Return a GeoHash representation of the geometry.'''),
    ('ST_3DIntersects', None,
     '''Returns TRUE if the Geometries \"spatially intersect\" in 3D - only for points, linestrings, polygons, polyhedral surface (area).'''),
    ('ST_Contains', None,
     '''[geometry] Returns true if and only if no points of B lie in the exterior of A, and at least one point of the interior of B lies in the interior of A.\nOR\n[raster] Return true if no points of raster rastB lie in the exterior of raster rastA and at least one point of the interior of rastB lies in the interior of rastA.'''),
    ('ST_ContainsProperly', None,
     '''[geometry] Returns true if B intersects the interior of A but not the boundary (or exterior). A does not contain properly itself, but does contain itself.\nOR\n[raster] Return true if rastB intersects the interior of rastA but not the boundary or exterior of rastA.'''),
    ('ST_Covers', None,
     '''[geometry] Returns 1 (TRUE) if no point in Geometry B is outside Geometry A\nOR\n[raster] Return true if no points of raster rastB lie outside raster rastA.'''),
    ('ST_CoveredBy', None,
     '''[geometry] Returns 1 (TRUE) if no point in Geometry/Geography A is outside Geometry/Geography B\nOR\n[raster] Return true if no points of raster rastA lie outside raster rastB.'''),
    ('ST_Crosses', None,
     '''Returns TRUE if the supplied geometries have some, but not all, interior points in common.'''),
    ('ST_LineCrossingDirection', None,
     '''Given 2 linestrings, returns a number between -3 and 3 denoting what kind of crossing behavior. 0 is no crossing.'''),
    ('ST_Disjoint', None,
     '''[geometry] Returns TRUE if the Geometries do not \"spatially intersect\" - if they do not share any space together.\nOR\n[raster] Return true if raster rastA does not spatially intersect rastB.'''),
    ('ST_Equals', None,
     '''Returns true if the given geometries represent the same geometry. Directionality is ignored.'''),
    ('ST_Intersects', None,
     '''[geometry] Returns TRUE if the Geometries/Geography \"spatially intersect in 2D\" - (share any portion of space) and FALSE if they don't (they are Disjoint). For geography tolerance is 0.00001 meters (so any points that close are considered to intersect)\nOR\n[raster] Return true if raster rastA spatially intersects raster rastB.'''),
    ('ST_OrderingEquals', None,
     '''Returns true if the given geometries represent the same geometry and points are in the same directional order.'''),
    ('ST_Overlaps', None,
     '''[geometry] Returns TRUE if the Geometries share space, are of the same dimension, but are not completely contained by each other.\nOR\n[raster] Return true if raster rastA and rastB intersect but one does not completely contain the other.'''),
    ('ST_PointInsideCircle', None,
     '''Is the point geometry inside the circle defined by center_x, center_y, radius'''),
    ('ST_Relate', None,
     '''Returns true if this Geometry is spatially related to anotherGeometry, by testing for intersections between the Interior, Boundary and Exterior of the two geometries as specified by the values in the intersectionMatrixPattern. If no intersectionMatrixPattern is passed in, then returns the maximum intersectionMatrixPattern that relates the 2 geometries.'''),
    ('ST_RelateMatch', None,
     '''Returns true if intersectionMattrixPattern1 implies intersectionMatrixPattern2'''),
    ('ST_Touches', None,
     '''[geometry] Returns TRUE if the geometries have at least one point in common, but their interiors do not intersect.\nOR\n[raster] Return true if raster rastA and rastB have at least one point in common but their interiors do not intersect.'''),
    ('ST_Within', None,
     '''[geometry] Returns true if the geometry A is completely inside geometry B\nOR\n[raster] Return true if no points of raster rastA lie in the exterior of raster rastB and at least one point of the interior of rastA lies in the interior of rastB.'''),
    ('ST_3DDWithin', None,
     '''For 3d (z) geometry type Returns true if two geometries 3d distance is within number of units.'''),
    ('ST_3DDFullyWithin', None,
     '''Returns true if all of the 3D geometries are within the specified distance of one another.'''),
    ('ST_DFullyWithin', None,
     '''[geometry] Returns true if all of the geometries are within the specified distance of one another\nOR\n[raster] Return true if rasters rastA and rastB are fully within the specified distance of each other.'''),
    ('ST_DWithin', None,
     '''[geometry] Returns true if the geometries are within the specified distance of one another. For geometry units are in those of spatial reference and for geography units are in meters and measurement is defaulted to use_spheroid=true (measure around spheroid), for faster check, use_spheroid=false to measure along sphere.\nOR\n[raster] Return true if rasters rastA and rastB are within the specified distance of each other.'''),
    ('ST_Area', None,
     '''Returns the area of a polygonal geometry.'''),
    ('ST_Azimuth', None,
     '''Returns the north-based azimuth as the angle in radians measured clockwise from the vertical on pointA to pointB.'''),
    ('ST_Angle', None,
     '''Returns the angle between 3 points, or between 2 vectors (4 points or 2 lines).'''),
    ('ST_ClosestPoint', types.Geometry,
     '''Returns the 2D point on g1 that is closest to g2. This is the first point of the shortest line.'''),
    ('ST_3DClosestPoint', types.Geometry,
     '''Returns the 3D point on g1 that is closest to g2. This is the first point of the 3D shortest line.'''),
    ('ST_Distance', None,
     '''Returns the distance between two geometry or geography values.'''),
    ('ST_3DDistance', None,
     '''Returns the 3D cartesian minimum distance (based on spatial ref) between two geometries in projected units.'''),
    ('ST_DistanceSphere', None,
     '''Returns minimum distance in meters between two lon/lat geometries using a spherical earth model.'''),
    ('ST_DistanceSpheroid', None,
     '''Returns the minimum distance between two lon/lat geometries using a spheroidal earth model.'''),
    ('ST_FrechetDistance', None,
     '''Returns the Fréchet distance between two geometries.'''),
    ('ST_HausdorffDistance', None,
     '''Returns the Hausdorff distance between two geometries.'''),
    ('ST_Length', None,
     '''Returns the 2D length of a linear geometry.'''),
    ('ST_Length2D', None,
     '''Returns the 2D length of a linear geometry. Alias for ST_Length'''),
    ('ST_3DLength', None,
     '''Returns the 3D length of a linear geometry.'''),
    ('ST_LengthSpheroid', None,
     '''Returns the 2D or 3D length/perimeter of a lon/lat geometry on a spheroid.'''),
    ('ST_LongestLine', types.Geometry,
     '''Returns the 2D longest line between two geometries.'''),
    ('ST_3DLongestLine', types.Geometry,
     '''Returns the 3D longest line between two geometries'''),
    ('ST_MaxDistance', None,
     '''Returns the 2D largest distance between two geometries in projected units.'''),
    ('ST_3DMaxDistance', None,
     '''Returns the 3D cartesian maximum distance (based on spatial ref) between two geometries in projected units.'''),
    ('ST_MinimumClearance', None,
     '''Returns the minimum clearance of a geometry, a measure of a geometry's robustness.'''),
    ('ST_MinimumClearanceLine', types.Geometry,
     '''Returns the two-point LineString spanning a geometry's minimum clearance.'''),
    ('ST_Perimeter', None,
     '''Returns the length of the boundary of a polygonal geometry or geography.'''),
    ('ST_Perimeter2D', None,
     '''Returns the 2D perimeter of a polygonal geometry. Alias for ST_Perimeter.'''),
    ('ST_3DPerimeter', None,
     '''Returns the 3D perimeter of a polygonal geometry.'''),
    ('ST_Project', types.Geography,
     '''Returns a point projected from a start point by a distance and bearing (azimuth).'''),
    ('ST_ShortestLine', types.Geometry,
     '''Returns the 2D shortest line between two geometries'''),
    ('ST_3DShortestLine', types.Geometry,
     '''Returns the 3D shortest line between two geometries'''),
    ('ST_Buffer', types.Geometry,
     '''(T) Returns a geometry covering all points within a given distance from the input geometry.'''),
    ('ST_BuildArea', types.Geometry,
     '''Creates an areal geometry formed by the constituent linework of given geometry'''),
    ('ST_Centroid', types.Geometry,
     '''Returns the geometric center of a geometry.'''),
    ('ST_ClipByBox2D', types.Geometry,
     '''Returns the portion of a geometry falling within a rectangle.'''),
    ('ST_ConcaveHull', types.Geometry,
     '''The concave hull of a geometry represents a possibly concave geometry that encloses all geometries within the set. You can think of it as shrink wrapping.'''),
    ('ST_ConvexHull', types.Geometry,
     '''[geometry] Computes the convex hull of a geometry.\nOR\n[raster] Return the convex hull geometry of the raster including pixel values equal to BandNoDataValue. For regular shaped and non-skewed rasters, this gives the same result as ST_Envelope so only useful for irregularly shaped or skewed rasters.'''),
    ('ST_CurveToLine', types.Geometry,
     '''Converts a CIRCULARSTRING/CURVEPOLYGON/MULTISURFACE to a LINESTRING/POLYGON/MULTIPOLYGON'''),
    ('ST_DelaunayTriangles', types.Geometry,
     '''Return a Delaunay triangulation around the given input points.'''),
    ('ST_Difference', types.Geometry,
     '''Returns a geometry that represents that part of geometry A that does not intersect with geometry B.'''),
    ('ST_FlipCoordinates', types.Geometry,
     '''Returns a version of the given geometry with X and Y axis flipped. Useful for people who have built latitude/longitude features and need to fix them.'''),
    ('ST_GeneratePoints', types.Geometry,
     '''Converts a polygon or multi-polygon into a multi-point composed of randomly location points within the original areas.'''),
    ('ST_GeometricMedian', types.Geometry,
     '''Returns the geometric median of a MultiPoint.'''),
    ('ST_Intersection', types.Geometry,
     '''[geometry] (T) Returns a geometry that represents the shared portion of geomA and geomB.\nOR\n[raster] Returns a raster or a set of geometry-pixelvalue pairs representing the shared portion of two rasters or the geometrical intersection of a vectorization of the raster and a geometry.'''),
    ('ST_LineToCurve', types.Geometry,
     '''Converts a LINESTRING/POLYGON to a CIRCULARSTRING, CURVEPOLYGON'''),
    ('ST_MakeValid', types.Geometry,
     '''Attempts to make an invalid geometry valid without losing vertices.'''),
    ('ST_MemUnion', types.Geometry,
     '''Same as ST_Union, only memory-friendly (uses less memory and more processor time).'''),
    ('ST_MinimumBoundingCircle', types.Geometry,
     '''Returns the smallest circle polygon that can fully contain a geometry. Default uses 48 segments per quarter circle.'''),
    ('ST_MinimumBoundingRadius', None,
     '''Returns the center point and radius of the smallest circle that can fully contain a geometry.'''),
    ('ST_OrientedEnvelope', types.Geometry,
     '''Returns a minimum rotated rectangle enclosing a geometry.'''),
    ('ST_Polygonize', types.Geometry,
     '''Aggregate. Creates a GeometryCollection containing possible polygons formed from the constituent linework of a set of geometries.'''),
    ('ST_Node', types.Geometry,
     '''Node a set of linestrings.'''),
    ('ST_OffsetCurve', types.Geometry,
     '''Return an offset line at a given distance and side from an input line. Useful for computing parallel lines about a center line'''),
    ('ST_PointOnSurface', types.Geometry,
     '''Returns a POINT guaranteed to lie on the surface.'''),
    ('ST_RemoveRepeatedPoints', types.Geometry,
     '''Returns a version of the given geometry with duplicated points removed.'''),
    ('ST_SharedPaths', types.Geometry,
     '''Returns a collection containing paths shared by the two input linestrings/multilinestrings.'''),
    ('ST_ShiftLongitude', types.Geometry,
     ('''Toggle geometry coordinates between -180..180 and 0..360 ranges.''', 'ST_Shift_Longitude')),
    ('ST_WrapX', types.Geometry,
     '''Wrap a geometry around an X value.'''),
    ('ST_Simplify', types.Geometry,
     '''Returns a \"simplified\" version of the given geometry using the Douglas-Peucker algorithm.'''),
    ('ST_SimplifyPreserveTopology', types.Geometry,
     '''Returns a \"simplified\" version of the given geometry using the Douglas-Peucker algorithm. Will avoid creating derived geometries (polygons in particular) that are invalid.'''),
    ('ST_SimplifyVW', types.Geometry,
     '''Returns a \"simplified\" version of the given geometry using the Visvalingam-Whyatt algorithm'''),
    ('ST_ChaikinSmoothing', types.Geometry,
     '''Returns a \"smoothed\" version of the given geometry using the Chaikin algorithm'''),
    ('ST_FilterByM', types.Geometry,
     '''Filters vertex points based on their m-value'''),
    ('ST_SetEffectiveArea', types.Geometry,
     '''Sets the effective area for each vertex, storing the value in the M ordinate. A simplified geometry can then be generated by filtering on the M ordinate.'''),
    ('ST_Split', types.Geometry,
     '''Returns a collection of geometries resulting by splitting a geometry.'''),
    ('ST_SymDifference', types.Geometry,
     '''Returns a geometry that represents the portions of A and B that do not intersect. It is called a symmetric difference because ST_SymDifference(A,B) = ST_SymDifference(B,A).'''),
    ('ST_Subdivide', types.Geometry,
     '''Returns a set of geometry where no geometry in the set has more than the specified number of vertices.'''),
    ('ST_Union', types.Geometry,
     '''[geometry] Returns a geometry that represents the point set union of the Geometries.\nOR\n[raster] Returns the union of a set of raster tiles into a single raster composed of 1 or more bands.'''),
    ('ST_UnaryUnion', types.Geometry,
     '''Like ST_Union, but working at the geometry component level.'''),
    ('ST_VoronoiLines', types.Geometry,
     '''Returns the boundaries between the cells of the Voronoi diagram constructed from the vertices of a geometry.'''),
    ('ST_VoronoiPolygons', types.Geometry,
     '''Returns the cells of the Voronoi diagram constructed from the vertices of a geometry.'''),
    ('ST_Affine', types.Geometry,
     '''Apply a 3D affine transformation to a geometry.'''),
    ('ST_Rotate', types.Geometry,
     '''Rotates a geometry about an origin point.'''),
    ('ST_RotateX', types.Geometry,
     '''Rotates a geometry about the X axis.'''),
    ('ST_RotateY', types.Geometry,
     '''Rotates a geometry about the Y axis.'''),
    ('ST_RotateZ', types.Geometry,
     '''Rotates a geometry about the Z axis.'''),
    ('ST_Scale', types.Geometry,
     '''Scales a geometry by given factors.'''),
    ('ST_Translate', types.Geometry,
     '''Translates a geometry by given offsets.'''),
    ('ST_TransScale', types.Geometry,
     '''Translates and scales a geometry by given offsets and factors.'''),
    ('ST_ClusterDBSCAN', None,
     '''Window function that returns a cluster id for each input geometry using the DBSCAN algorithm.'''),
    ('ST_ClusterIntersecting', types.Geometry,
     '''Aggregate function that clusters the input geometries into connected sets.'''),
    ('ST_ClusterKMeans', None,
     '''Window function that returns a cluster id for each input geometry using the K-means algorithm.'''),
    ('ST_ClusterWithin', types.Geometry,
     '''Aggregate function that clusters the input geometries by separation distance.'''),
    ('Box2D', None,  # return an unsupported Box2d object
     ('''Returns a BOX2D representing the 2D extent of the geometry.''', 'Box2D_type')),
    ('Box3D', None,  # return an unsupported Box3d object
     ('''[geometry] Returns a BOX3D representing the 3D extent of the geometry.\nOR\n[raster] Returns the box 3d representation of the enclosing box of the raster.''', 'Box3D_type')),
    ('ST_EstimatedExtent', None,  # return an unsupported Box2d object
     '''Return the 'estimated' extent of a spatial table.'''),
    ('ST_Expand', types.Geometry,
     '''Returns a bounding box expanded from another bounding box or a geometry.'''),
    ('ST_Extent', None,  # return an unsupported Box2d object
     '''an aggregate function that returns the bounding box that bounds rows of geometries.'''),
    ('ST_3DExtent', None,  # return an unsupported Box3d object
     '''an aggregate function that returns the 3D bounding box that bounds rows of geometries.'''),
    ('ST_MakeBox2D', None,  # return an unsupported Box2d object
     '''Creates a BOX2D defined by two 2D point geometries.'''),
    ('ST_3DMakeBox', None,  # return an unsupported Box3d object
     '''Creates a BOX3D defined by two 3D point geometries.'''),
    ('ST_XMax', None,
     '''Returns the X maxima of a 2D or 3D bounding box or a geometry.'''),
    ('ST_XMin', None,
     '''Returns the X minima of a 2D or 3D bounding box or a geometry.'''),
    ('ST_YMax', None,
     '''Returns the Y maxima of a 2D or 3D bounding box or a geometry.'''),
    ('ST_YMin', None,
     '''Returns the Y minima of a 2D or 3D bounding box or a geometry.'''),
    ('ST_ZMax', None,
     '''Returns the Z maxima of a 2D or 3D bounding box or a geometry.'''),
    ('ST_ZMin', None,
     '''Returns the Z minima of a 2D or 3D bounding box or a geometry.'''),
    ('ST_LineInterpolatePoint', types.Geometry,
     '''Returns a point interpolated along a line. Second argument is a float8 between 0 and 1 representing fraction of total length of linestring the point has to be located.'''),
    ('ST_3DLineInterpolatePoint', types.Geometry,
     '''Returns a point interpolated along a line in 3D. Second argument is a float8 between 0 and 1 representing fraction of total length of linestring the point has to be located.'''),
    ('ST_LineInterpolatePoints', types.Geometry,
     '''Returns one or more points interpolated along a line.'''),
    ('ST_LineLocatePoint', None,
     '''Returns a float between 0 and 1 representing the location of the closest point on LineString to the given Point, as a fraction of total 2d line length.'''),
    ('ST_LineSubstring', types.Geometry,
     '''Return a linestring being a substring of the input one starting and ending at the given fractions of total 2d length. Second and third arguments are float8 values between 0 and 1.'''),
    ('ST_LocateAlong', types.Geometry,
     '''Return a derived geometry collection value with elements that match the specified measure. Polygonal elements are not supported.'''),
    ('ST_LocateBetween', types.Geometry,
     '''Return a derived geometry collection value with elements that match the specified range of measures inclusively.'''),
    ('ST_LocateBetweenElevations', types.Geometry,
     '''Return a derived geometry (collection) value with elements that intersect the specified range of elevations inclusively.'''),
    ('ST_InterpolatePoint', None,
     '''Return the value of the measure dimension of a geometry at the point closed to the provided point.'''),
    ('ST_AddMeasure', types.Geometry,
     '''Return a derived geometry with measure elements linearly interpolated between the start and end points.'''),
    ('ST_IsValidTrajectory', None,
     '''Returns true if the geometry is a valid trajectory.'''),
    ('ST_ClosestPointOfApproach', None,
     '''Returns the measure at which points interpolated along two trajectories are closest.'''),
    ('ST_DistanceCPA', None,
     '''Returns the distance between the closest point of approach of two trajectories.'''),
    ('ST_CPAWithin', None,
     '''Returns true if the closest point of approach of two trajectories is within the specified distance.'''),
    ('postgis_sfcgal_version', None,
     '''Returns the version of SFCGAL in use'''),
    ('ST_Extrude', types.Geometry,
     '''Extrude a surface to a related volume'''),
    ('ST_StraightSkeleton', types.Geometry,
     '''Compute a straight skeleton from a geometry'''),
    ('ST_ApproximateMedialAxis', types.Geometry,
     '''Compute the approximate medial axis of an areal geometry.'''),
    ('ST_IsPlanar', None,
     '''Check if a surface is or not planar'''),
    ('ST_Orientation', None,
     '''Determine surface orientation'''),
    ('ST_ForceLHR', types.Geometry,
     '''Force LHR orientation'''),
    ('ST_MinkowskiSum', types.Geometry,
     '''Performs Minkowski sum'''),
    ('ST_ConstrainedDelaunayTriangles', types.Geometry,
     '''Return a constrained Delaunay triangulation around the given input geometry.'''),
    ('ST_3DIntersection', types.Geometry,
     '''Perform 3D intersection'''),
    ('ST_3DDifference', types.Geometry,
     '''Perform 3D difference'''),
    ('ST_3DUnion', types.Geometry,
     '''Perform 3D union'''),
    ('ST_3DArea', None,
     '''Computes area of 3D surface geometries. Will return 0 for solids.'''),
    ('ST_Tesselate', types.Geometry,
     '''Perform surface Tessellation of a polygon or polyhedralsurface and returns as a TIN or collection of TINS'''),
    ('ST_Volume', None,
     '''Computes the volume of a 3D solid. If applied to surface (even closed) geometries will return 0.'''),
    ('ST_MakeSolid', types.Geometry,
     '''Cast the geometry into a solid. No check is performed. To obtain a valid solid, the input geometry must be a closed Polyhedral Surface or a closed TIN.'''),
    ('ST_IsSolid', None,
     '''Test if the geometry is a solid. No validity check is performed.'''),
    ('AddAuth', None,
     '''Adds an authorization token to be used in the current transaction.'''),
    ('CheckAuth', None,
     '''Creates a trigger on a table to prevent/allow updates and deletes of rows based on authorization token.'''),
    ('DisableLongTransactions', None,
     '''Disables long transaction support.'''),
    ('EnableLongTransactions', None,
     '''Enables long transaction support.'''),
    ('LockRow', None,
     '''Sets lock/authorization for a row in a table.'''),
    ('UnlockRows', None,
     '''Removes all locks held by an authorization token.'''),
    ('PostGIS_Extensions_Upgrade', None,
     '''Packages and upgrades postgis extensions (e.g. postgis_raster, postgis_topology, postgis_sfcgal) to latest available version.'''),
    ('PostGIS_Full_Version', None,
     '''Reports full postgis version and build configuration infos.'''),
    ('PostGIS_GEOS_Version', None,
     '''Returns the version number of the GEOS library.'''),
    ('PostGIS_Liblwgeom_Version', None,
     '''Returns the version number of the liblwgeom library. This should match the version of PostGIS.'''),
    ('PostGIS_LibXML_Version', None,
     '''Returns the version number of the libxml2 library.'''),
    ('PostGIS_Lib_Build_Date', None,
     '''Returns build date of the PostGIS library.'''),
    ('PostGIS_Lib_Version', None,
     '''Returns the version number of the PostGIS library.'''),
    ('PostGIS_PROJ_Version', None,
     '''Returns the version number of the PROJ4 library.'''),
    ('PostGIS_Wagyu_Version', None,
     '''Returns the version number of the internal Wagyu library.'''),
    ('PostGIS_Scripts_Build_Date', None,
     '''Returns build date of the PostGIS scripts.'''),
    ('PostGIS_Scripts_Installed', None,
     '''Returns version of the postgis scripts installed in this database.'''),
    ('PostGIS_Scripts_Released', None,
     '''Returns the version number of the postgis.sql script released with the installed postgis lib.'''),
    ('PostGIS_Version', None,
     '''Returns PostGIS version number and compile-time options.'''),
    ('PostGIS_AddBBox', types.Geometry,
     '''Add bounding box to the geometry.'''),
    ('PostGIS_DropBBox', types.Geometry,
     '''Drop the bounding box cache from the geometry.'''),
    ('PostGIS_HasBBox', None,
     '''Returns TRUE if the bbox of this geometry is cached, FALSE otherwise.'''),
    ('ST_AddBand', types.Raster,
     ('''Returns a raster with the new band(s) of given type added with given initial value in the given index location. If no index is specified, the band is added to the end.''', 'RT_ST_AddBand')),
    ('ST_AsRaster', types.Raster,
     ('''Converts a PostGIS geometry to a PostGIS raster.''', 'RT_ST_AsRaster')),
    ('ST_Band', types.Raster,
     ('''Returns one or more bands of an existing raster as a new raster. Useful for building new rasters from existing rasters.''', 'RT_ST_Band')),
    ('ST_MakeEmptyCoverage', types.Raster,
     ('''Cover georeferenced area with a grid of empty raster tiles.''', 'RT_ST_MakeEmptyCoverage')),
    ('ST_MakeEmptyRaster', types.Raster,
     ('''Returns an empty raster (having no bands) of given dimensions (width & height), upperleft X and Y, pixel size and rotation (scalex, scaley, skewx & skewy) and reference system (srid). If a raster is passed in, returns a new raster with the same size, alignment and SRID. If srid is left out, the spatial ref is set to unknown (0).''', 'RT_ST_MakeEmptyRaster')),
    ('ST_Tile', types.Raster,
     ('''Returns a set of rasters resulting from the split of the input raster based upon the desired dimensions of the output rasters.''', 'RT_ST_Tile')),
    ('ST_Retile', types.Raster,
     ('''Return a set of configured tiles from an arbitrarily tiled raster coverage.''', 'RT_ST_Retile')),
    ('ST_FromGDALRaster', types.Raster,
     ('''Returns a raster from a supported GDAL raster file.''', 'RT_ST_FromGDALRaster')),
    ('ST_GeoReference', None,
     ('''Returns the georeference meta data in GDAL or ESRI format as commonly seen in a world file. Default is GDAL.''', 'RT_ST_GeoReference')),
    ('ST_Height', None,
     ('''Returns the height of the raster in pixels.''', 'RT_ST_Height')),
    ('ST_MetaData', None,
     ('''Returns basic meta data about a raster object such as pixel size, rotation (skew), upper, lower left, etc.''', 'RT_ST_MetaData')),
    ('ST_NumBands', None,
     ('''Returns the number of bands in the raster object.''', 'RT_ST_NumBands')),
    ('ST_PixelHeight', None,
     ('''Returns the pixel height in geometric units of the spatial reference system.''', 'RT_ST_PixelHeight')),
    ('ST_PixelWidth', None,
     ('''Returns the pixel width in geometric units of the spatial reference system.''', 'RT_ST_PixelWidth')),
    ('ST_ScaleX', None,
     ('''Returns the X component of the pixel width in units of coordinate reference system.''', 'RT_ST_ScaleX')),
    ('ST_ScaleY', None,
     ('''Returns the Y component of the pixel height in units of coordinate reference system.''', 'RT_ST_ScaleY')),
    ('ST_RasterToWorldCoord', None,
     ('''Returns the raster's upper left corner as geometric X and Y (longitude and latitude) given a column and row. Column and row starts at 1.''', 'RT_ST_RasterToWorldCoord')),
    ('ST_RasterToWorldCoordX', None,
     ('''Returns the geometric X coordinate upper left of a raster, column and row. Numbering of columns and rows starts at 1.''', 'RT_ST_RasterToWorldCoordX')),
    ('ST_RasterToWorldCoordY', None,
     ('''Returns the geometric Y coordinate upper left corner of a raster, column and row. Numbering of columns and rows starts at 1.''', 'RT_ST_RasterToWorldCoordY')),
    ('ST_Rotation', None,
     ('''Returns the rotation of the raster in radian.''', 'RT_ST_Rotation')),
    ('ST_SkewX', None,
     ('''Returns the georeference X skew (or rotation parameter).''', 'RT_ST_SkewX')),
    ('ST_SkewY', None,
     ('''Returns the georeference Y skew (or rotation parameter).''', 'RT_ST_SkewY')),
    ('ST_UpperLeftX', None,
     ('''Returns the upper left X coordinate of raster in projected spatial ref.''', 'RT_ST_UpperLeftX')),
    ('ST_UpperLeftY', None,
     ('''Returns the upper left Y coordinate of raster in projected spatial ref.''', 'RT_ST_UpperLeftY')),
    ('ST_Width', None,
     ('''Returns the width of the raster in pixels.''', 'RT_ST_Width')),
    ('ST_WorldToRasterCoord', None,
     ('''Returns the upper left corner as column and row given geometric X and Y (longitude and latitude) or a point geometry expressed in the spatial reference coordinate system of the raster.''', 'RT_ST_WorldToRasterCoord')),
    ('ST_WorldToRasterCoordX', None,
     ('''Returns the column in the raster of the point geometry (pt) or a X and Y world coordinate (xw, yw) represented in world spatial reference system of raster.''', 'RT_ST_WorldToRasterCoordX')),
    ('ST_WorldToRasterCoordY', None,
     ('''Returns the row in the raster of the point geometry (pt) or a X and Y world coordinate (xw, yw) represented in world spatial reference system of raster.''', 'RT_ST_WorldToRasterCoordY')),
    ('ST_BandMetaData', None,
     ('''Returns basic meta data for a specific raster band. band num 1 is assumed if none-specified.''', 'RT_ST_BandMetaData')),
    ('ST_BandNoDataValue', None,
     ('''Returns the value in a given band that represents no data. If no band num 1 is assumed.''', 'RT_ST_BandNoDataValue')),
    ('ST_BandIsNoData', None,
     ('''Returns true if the band is filled with only nodata values.''', 'RT_ST_BandIsNoData')),
    ('ST_BandPath', None,
     ('''Returns system file path to a band stored in file system. If no bandnum specified, 1 is assumed.''', 'RT_ST_BandPath')),
    ('ST_BandFileSize', None,
     ('''Returns the file size of a band stored in file system. If no bandnum specified, 1 is assumed.''', 'RT_ST_BandFileSize')),
    ('ST_BandFileTimestamp', None,
     ('''Returns the file timestamp of a band stored in file system. If no bandnum specified, 1 is assumed.''', 'RT_ST_BandFileTimestamp')),
    ('ST_BandPixelType', None,
     ('''Returns the type of pixel for given band. If no bandnum specified, 1 is assumed.''', 'RT_ST_BandPixelType')),
    ('ST_MinPossibleValue', None,
     '''Returns the minimum value this pixeltype can store.'''),
    ('ST_HasNoBand', None,
     ('''Returns true if there is no band with given band number. If no band number is specified, then band number 1 is assumed.''', 'RT_ST_HasNoBand')),
    ('ST_PixelAsPolygon', types.Geometry,
     ('''Returns the polygon geometry that bounds the pixel for a particular row and column.''', 'RT_ST_PixelAsPolygon')),
    ('ST_PixelAsPolygons', None,
     ('''Returns the polygon geometry that bounds every pixel of a raster band along with the value, the X and the Y raster coordinates of each pixel.''', 'RT_ST_PixelAsPolygons')),
    ('ST_PixelAsPoint', types.Geometry,
     ('''Returns a point geometry of the pixel's upper-left corner.''', 'RT_ST_PixelAsPoint')),
    ('ST_PixelAsPoints', None,
     ('''Returns a point geometry for each pixel of a raster band along with the value, the X and the Y raster coordinates of each pixel. The coordinates of the point geometry are of the pixel's upper-left corner.''', 'RT_ST_PixelAsPoints')),
    ('ST_PixelAsCentroid', types.Geometry,
     ('''Returns the centroid (point geometry) of the area represented by a pixel.''', 'RT_ST_PixelAsCentroid')),
    ('ST_PixelAsCentroids', None,
     ('''Returns the centroid (point geometry) for each pixel of a raster band along with the value, the X and the Y raster coordinates of each pixel. The point geometry is the centroid of the area represented by a pixel.''', 'RT_ST_PixelAsCentroids')),
    ('ST_Value', None,
     ('''Returns the value of a given band in a given columnx, rowy pixel or at a particular geometric point. Band numbers start at 1 and assumed to be 1 if not specified. If exclude_nodata_value is set to false, then all pixels include nodata pixels are considered to intersect and return value. If exclude_nodata_value is not passed in then reads it from metadata of raster.''', 'RT_ST_Value')),
    ('ST_NearestValue', None,
     ('''Returns the nearest non-NODATA value of a given band's pixel specified by a columnx and rowy or a geometric point expressed in the same spatial reference coordinate system as the raster.''', 'RT_ST_NearestValue')),
    ('ST_Neighborhood', None,
     ('''Returns a 2-D double precision array of the non-NODATA values around a given band's pixel specified by either a columnX and rowY or a geometric point expressed in the same spatial reference coordinate system as the raster.''', 'RT_ST_Neighborhood')),
    ('ST_SetValue', types.Raster,
     ('''Returns modified raster resulting from setting the value of a given band in a given columnx, rowy pixel or the pixels that intersect a particular geometry. Band numbers start at 1 and assumed to be 1 if not specified.''', 'RT_ST_SetValue')),
    ('ST_SetValues', types.Raster,
     ('''Returns modified raster resulting from setting the values of a given band.''', 'RT_ST_SetValues')),
    ('ST_DumpValues', None,
     ('''Get the values of the specified band as a 2-dimension array.''', 'RT_ST_DumpValues')),
    ('ST_PixelOfValue', None,
     ('''Get the columnx, rowy coordinates of the pixel whose value equals the search value.''', 'RT_ST_PixelOfValue')),
    ('ST_SetGeoReference', types.Raster,
     ('''Set Georeference 6 georeference parameters in a single call. Numbers should be separated by white space. Accepts inputs in GDAL or ESRI format. Default is GDAL.''', 'RT_ST_SetGeoReference')),
    ('ST_SetRotation', types.Raster,
     ('''Set the rotation of the raster in radian.''', 'RT_ST_SetRotation')),
    ('ST_SetScale', types.Raster,
     ('''Sets the X and Y size of pixels in units of coordinate reference system. Number units/pixel width/height.''', 'RT_ST_SetScale')),
    ('ST_SetSkew', types.Raster,
     ('''Sets the georeference X and Y skew (or rotation parameter). If only one is passed in, sets X and Y to the same value.''', 'RT_ST_SetSkew')),
    ('ST_SetUpperLeft', types.Raster,
     ('''Sets the value of the upper left corner of the pixel of the raster to projected X and Y coordinates.''', 'RT_ST_SetUpperLeft')),
    ('ST_Resample', types.Raster,
     ('''Resample a raster using a specified resampling algorithm, new dimensions, an arbitrary grid corner and a set of raster georeferencing attributes defined or borrowed from another raster.''', 'RT_ST_Resample')),
    ('ST_Rescale', types.Raster,
     ('''Resample a raster by adjusting only its scale (or pixel size). New pixel values are computed using the NearestNeighbor (english or american spelling), Bilinear, Cubic, CubicSpline or Lanczos resampling algorithm. Default is NearestNeighbor.''', 'RT_ST_Rescale')),
    ('ST_Reskew', types.Raster,
     ('''Resample a raster by adjusting only its skew (or rotation parameters). New pixel values are computed using the NearestNeighbor (english or american spelling), Bilinear, Cubic, CubicSpline or Lanczos resampling algorithm. Default is NearestNeighbor.''', 'RT_ST_Reskew')),
    ('ST_Resize', types.Raster,
     ('''Resize a raster to a new width/height''', 'RT_ST_Resize')),
    ('ST_SetBandNoDataValue', types.Raster,
     ('''Sets the value for the given band that represents no data. Band 1 is assumed if no band is specified. To mark a band as having no nodata value, set the nodata value = NULL.''', 'RT_ST_SetBandNoDataValue')),
    ('ST_SetBandIsNoData', types.Raster,
     ('''Sets the isnodata flag of the band to TRUE.''', 'RT_ST_SetBandIsNoData')),
    ('ST_SetBandPath', types.Raster,
     ('''Update the external path and band number of an out-db band''', 'RT_ST_SetBandPath')),
    ('ST_SetBandIndex', types.Raster,
     ('''Update the external band number of an out-db band''', 'RT_ST_SetBandIndex')),
    ('ST_Count', None,
     ('''Returns the number of pixels in a given band of a raster or raster coverage. If no band is specified defaults to band 1. If exclude_nodata_value is set to true, will only count pixels that are not equal to the nodata value.''', 'RT_ST_Count')),
    ('ST_CountAgg', None,
     ('''Aggregate. Returns the number of pixels in a given band of a set of rasters. If no band is specified defaults to band 1. If exclude_nodata_value is set to true, will only count pixels that are not equal to the NODATA value.''', 'RT_ST_CountAgg')),
    ('ST_Histogram', None,
     ('''Returns a set of record summarizing a raster or raster coverage data distribution separate bin ranges. Number of bins are autocomputed if not specified.''', 'RT_ST_Histogram')),
    ('ST_Quantile', None,
     ('''Compute quantiles for a raster or raster table coverage in the context of the sample or population. Thus, a value could be examined to be at the raster's 25%, 50%, 75% percentile.''', 'RT_ST_Quantile')),
    ('ST_SummaryStats', None,
     ('''Returns summarystats consisting of count, sum, mean, stddev, min, max for a given raster band of a raster or raster coverage. Band 1 is assumed is no band is specified.''', 'RT_ST_SummaryStats')),
    ('ST_SummaryStatsAgg', types.SummaryStats,
     ('''Aggregate. Returns summarystats consisting of count, sum, mean, stddev, min, max for a given raster band of a set of raster. Band 1 is assumed is no band is specified.''', 'RT_ST_SummaryStatsAgg')),
    ('ST_ValueCount', None,
     ('''Returns a set of records containing a pixel band value and count of the number of pixels in a given band of a raster (or a raster coverage) that have a given set of values. If no band is specified defaults to band 1. By default nodata value pixels are not counted. and all other values in the pixel are output and pixel band values are rounded to the nearest integer.''', 'RT_ST_ValueCount')),
    ('ST_RastFromWKB', types.Raster,
     ('''Return a raster value from a Well-Known Binary (WKB) raster.''', 'RT_ST_RastFromWKB')),
    ('ST_RastFromHexWKB', types.Raster,
     ('''Return a raster value from a Hex representation of Well-Known Binary (WKB) raster.''', 'RT_ST_RastFromHexWKB')),
    ('ST_AsWKB', None,
     ('''Return the Well-Known Binary (WKB) representation of the raster.''', 'RT_ST_AsBinary')),
    ('ST_AsHexWKB', None,
     ('''Return the Well-Known Binary (WKB) in Hex representation of the raster.''', 'RT_ST_AsHexWKB')),
    ('ST_AsGDALRaster', None,
     ('''Return the raster tile in the designated GDAL Raster format. Raster formats are one of those supported by your compiled library. Use ST_GDALDrivers() to get a list of formats supported by your library.''', 'RT_ST_AsGDALRaster')),
    ('ST_AsJPEG', None,
     ('''Return the raster tile selected bands as a single Joint Photographic Exports Group (JPEG) image (byte array). If no band is specified and 1 or more than 3 bands, then only the first band is used. If only 3 bands then all 3 bands are used and mapped to RGB.''', 'RT_ST_AsJPEG')),
    ('ST_AsPNG', None,
     ('''Return the raster tile selected bands as a single portable network graphics (PNG) image (byte array). If 1, 3, or 4 bands in raster and no bands are specified, then all bands are used. If more 2 or more than 4 bands and no bands specified, then only band 1 is used. Bands are mapped to RGB or RGBA space.''', 'RT_ST_AsPNG')),
    ('ST_AsTIFF', None,
     ('''Return the raster selected bands as a single TIFF image (byte array). If no band is specified or any of specified bands does not exist in the raster, then will try to use all bands.''', 'RT_ST_AsTIFF')),
    ('ST_Clip', types.Raster,
     ('''Returns the raster clipped by the input geometry. If band number not is specified, all bands are processed. If crop is not specified or TRUE, the output raster is cropped.''', 'RT_ST_Clip')),
    ('ST_ColorMap', types.Raster,
     ('''Creates a new raster of up to four 8BUI bands (grayscale, RGB, RGBA) from the source raster and a specified band. Band 1 is assumed if not specified.''', 'RT_ST_ColorMap')),
    ('ST_Grayscale', types.Raster,
     ('''Creates a new one-8BUI band raster from the source raster and specified bands representing Red, Green and Blue''', 'RT_ST_Grayscale')),
    ('ST_MapAlgebra', None,
     ('''[raster] Callback function version - Returns a one-band raster given one or more input rasters, band indexes and one user-specified callback function.\nOR\n[raster] Expression version - Returns a one-band raster given one or two input rasters, band indexes and one or more user-specified SQL expressions.''', 'RT_ST_MapAlgebra')),
    ('ST_MapAlgebraExpr', types.Raster,
     ('''[raster] 1 raster band version: Creates a new one band raster formed by applying a valid PostgreSQL algebraic operation on the input raster band and of pixeltype provided. Band 1 is assumed if no band is specified.\nOR\n[raster] 2 raster band version: Creates a new one band raster formed by applying a valid PostgreSQL algebraic operation on the two input raster bands and of pixeltype provided. band 1 of each raster is assumed if no band numbers are specified. The resulting raster will be aligned (scale, skew and pixel corners) on the grid defined by the first raster and have its extent defined by the \"extenttype\" parameter. Values for \"extenttype\" can be: INTERSECTION, UNION, FIRST, SECOND.''', 'RT_ST_MapAlgebraExpr')),
    ('ST_MapAlgebraFct', types.Raster,
     ('''[raster] 1 band version - Creates a new one band raster formed by applying a valid PostgreSQL function on the input raster band and of pixeltype prodived. Band 1 is assumed if no band is specified.\nOR\n[raster] 2 band version - Creates a new one band raster formed by applying a valid PostgreSQL function on the 2 input raster bands and of pixeltype prodived. Band 1 is assumed if no band is specified. Extent type defaults to INTERSECTION if not specified.''', 'RT_ST_MapAlgebraFct')),
    ('ST_MapAlgebraFctNgb', types.Raster,
     ('''1-band version: Map Algebra Nearest Neighbor using user-defined PostgreSQL function. Return a raster which values are the result of a PLPGSQL user function involving a neighborhood of values from the input raster band.''', 'RT_ST_MapAlgebraFctNgb')),
    ('ST_Reclass', types.Raster,
     ('''Creates a new raster composed of band types reclassified from original. The nband is the band to be changed. If nband is not specified assumed to be 1. All other bands are returned unchanged. Use case: convert a 16BUI band to a 8BUI and so forth for simpler rendering as viewable formats.''', 'RT_ST_Reclass')),
    ('ST_Distinct4ma', None,
     ('''Raster processing function that calculates the number of unique pixel values in a neighborhood.''', 'RT_ST_Distinct4ma')),
    ('ST_InvDistWeight4ma', None,
     ('''Raster processing function that interpolates a pixel's value from the pixel's neighborhood.''', 'RT_ST_InvDistWeight4ma')),
    ('ST_Max4ma', None,
     ('''Raster processing function that calculates the maximum pixel value in a neighborhood.''', 'RT_ST_Max4ma')),
    ('ST_Mean4ma', None,
     ('''Raster processing function that calculates the mean pixel value in a neighborhood.''', 'RT_ST_Mean4ma')),
    ('ST_Min4ma', None,
     ('''Raster processing function that calculates the minimum pixel value in a neighborhood.''', 'RT_ST_Min4ma')),
    ('ST_MinDist4ma', None,
     ('''Raster processing function that returns the minimum distance (in number of pixels) between the pixel of interest and a neighboring pixel with value.''', 'RT_ST_MinDist4ma')),
    ('ST_Range4ma', None,
     ('''Raster processing function that calculates the range of pixel values in a neighborhood.''', 'RT_ST_Range4ma')),
    ('ST_StdDev4ma', None,
     ('''Raster processing function that calculates the standard deviation of pixel values in a neighborhood.''', 'RT_ST_StdDev4ma')),
    ('ST_Sum4ma', None,
     ('''Raster processing function that calculates the sum of all pixel values in a neighborhood.''', 'RT_ST_Sum4ma')),
    ('ST_Aspect', types.Raster,
     ('''Returns the aspect (in degrees by default) of an elevation raster band. Useful for analyzing terrain.''', 'RT_ST_Aspect')),
    ('ST_HillShade', types.Raster,
     ('''Returns the hypothetical illumination of an elevation raster band using provided azimuth, altitude, brightness and scale inputs.''', 'RT_ST_HillShade')),
    ('ST_Roughness', types.Raster,
     ('''Returns a raster with the calculated \"roughness\" of a DEM.''', 'RT_ST_Roughness')),
    ('ST_Slope', types.Raster,
     ('''Returns the slope (in degrees by default) of an elevation raster band. Useful for analyzing terrain.''', 'RT_ST_Slope')),
    ('ST_TPI', types.Raster,
     ('''Returns a raster with the calculated Topographic Position Index.''', 'RT_ST_TPI')),
    ('ST_TRI', types.Raster,
     ('''Returns a raster with the calculated Terrain Ruggedness Index.''', 'RT_ST_TRI')),
    ('ST_DumpAsPolygons', None,
     ('''Returns a set of geomval (geom,val) rows, from a given raster band. If no band number is specified, band num defaults to 1.''', 'RT_ST_DumpAsPolygons')),
    ('ST_MinConvexHull', types.Geometry,
     ('''Return the convex hull geometry of the raster excluding NODATA pixels.''', 'RT_ST_MinConvexHull')),
    ('ST_SameAlignment', None,
     ('''Returns true if rasters have same skew, scale, spatial ref, and offset (pixels can be put on same grid without cutting into pixels) and false if they don't with notice detailing issue.''', 'RT_ST_SameAlignment')),
    ('ST_NotSameAlignmentReason', None,
     ('''Returns text stating if rasters are aligned and if not aligned, a reason why.''', 'RT_ST_NotSameAlignmentReason')),
    ('ST_Distance_Sphere', None,
     '''Returns minimum distance in meters between two lon/lat geometries. Uses a spherical earth and radius of 6370986 meters. Faster than ``ST_Distance_Spheroid``, but less accurate. PostGIS versions prior to 1.5 only implemented for points.'''),
]



"""Some helpers to use with Alembic migration tool."""
from alembic.autogenerate import renderers
from alembic.autogenerate import rewriter
from alembic.autogenerate.render import _add_column
from alembic.autogenerate.render import _add_index
from alembic.autogenerate.render import _add_table
from alembic.autogenerate.render import _drop_column
from alembic.autogenerate.render import _drop_index
from alembic.autogenerate.render import _drop_table
from alembic.ddl.base import RenameTable
from alembic.ddl.base import format_table_name
from alembic.ddl.base import visit_rename_table
from alembic.operations import BatchOperations
from alembic.operations import Operations
from alembic.operations import ops
from sqlalchemy import text
from sqlalchemy.dialects.sqlite.base import SQLiteDialect
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.schema import DropTable
from sqlalchemy.types import TypeDecorator

from geoalchemy2 import Column
from geoalchemy2 import Geography
from geoalchemy2 import Geometry
from geoalchemy2 import Raster
from geoalchemy2 import _check_spatial_type
from geoalchemy2 import _get_gis_cols
from geoalchemy2 import _spatial_idx_name
from geoalchemy2 import func

writer = rewriter.Rewriter()
"""Rewriter object for Alembic."""

_SPATIAL_TABLES = set()


def _monkey_patch_get_indexes_for_sqlite():
    """Monkey patch SQLAlchemy to fix spatial index reflection."""
    normal_behavior = SQLiteDialect.get_indexes

    def spatial_behavior(self, connection, table_name, schema=None, **kw):
        indexes = self._get_indexes_normal_behavior(
            connection, table_name, schema=None, **kw
        )

        # Check that SpatiaLite was loaded into the DB
        is_spatial_db = connection.exec_driver_sql(
            """PRAGMA main.table_info(geometry_columns)"""
        ).fetchall()
        if not is_spatial_db:
            return indexes

        # Get spatial indexes
        spatial_index_query = text(
            """SELECT *
            FROM geometry_columns
            WHERE f_table_name = '{}'
            ORDER BY f_table_name, f_geometry_column;""".format(
                table_name
            )
        )
        spatial_indexes = connection.execute(spatial_index_query).fetchall()
        if spatial_indexes:
            reflected_names = set([i["name"] for i in indexes])
            for idx in spatial_indexes:
                idx_col = idx[1]
                idx_name = _spatial_idx_name(table_name, idx_col)
                if not bool(idx[-1]) or idx_name in reflected_names:
                    continue
                indexes.append(
                    {
                        "name": idx_name,
                        "column_names": [idx_col],
                        "unique": 0,
                        "dialect_options": {"_column_flag": True},
                    }
                )
                reflected_names.add(idx_name)

        return indexes

    spatial_behavior.__doc__ = normal_behavior.__doc__
    SQLiteDialect.get_indexes = spatial_behavior
    SQLiteDialect._get_indexes_normal_behavior = normal_behavior


_monkey_patch_get_indexes_for_sqlite()


def render_item(obj_type, obj, autogen_context):
    """Add proper imports for spatial types."""
    if obj_type == "type" and isinstance(obj, (Geometry, Geography, Raster)):
        import_name = obj.__class__.__name__
        autogen_context.imports.add(f"from geoalchemy2 import {import_name}")
        return "%r" % obj

    # Default rendering for other objects
    return False


def include_object(obj, name, obj_type, reflected, compare_to):
    """Do not include internal tables of spatial extensions.

    .. warning::
        This function only checks the table names, so it might exclude tables that should not be.
        In such case, you should create your own function to handle your specific table names.

    """
    if (
        obj_type == "table"
        and (
            name.startswith("geometry_columns")
            or name.startswith("spatial_ref_sys")
            or name.startswith("spatialite_history")
            or name.startswith("sqlite_sequence")
            or name.startswith("views_geometry_columns")
            or name.startswith("virts_geometry_columns")
            or name.startswith("idx_")
        )
    ):
        return False
    return True


@Operations.register_operation("add_geospatial_column")
@BatchOperations.register_operation(
    "add_geospatial_column", "batch_add_geospatial_column"
)
class AddGeospatialColumnOp(ops.AddColumnOp):
    """
    Add a Geospatial Column in an Alembic migration context. This methodology originates from:
    https://alembic.sqlalchemy.org/en/latest/api/operations.html#operation-plugins
    """

    @classmethod
    def add_geospatial_column(cls, operations, table_name, column, schema=None):
        """Handle the different situations arising from adding geospatial column to a DB."""
        op = cls(table_name, column, schema=schema)
        return operations.invoke(op)

    def reverse(self):
        """Used to autogenerate the downgrade function."""
        return DropGeospatialColumnOp.from_column_and_tablename(
            self.schema, self.table_name, self.column.name
        )

    @classmethod
    def batch_add_geospatial_column(
        cls,
        operations,
        column,
        insert_before=None,
        insert_after=None,
    ):
        """Issue an "add column" instruction using the current batch migration context."""
        kw = {}
        if insert_before:
            kw["insert_before"] = insert_before
        if insert_after:
            kw["insert_after"] = insert_after

        op = cls(
            operations.impl.table_name,
            column,
            schema=operations.impl.schema,
            **kw,
        )
        return operations.invoke(op)


@Operations.register_operation("drop_geospatial_column")
@BatchOperations.register_operation(
    "drop_geospatial_column", "batch_drop_geospatial_column"
)
class DropGeospatialColumnOp(ops.DropColumnOp):
    """Drop a Geospatial Column in an Alembic migration context."""

    @classmethod
    def drop_geospatial_column(
        cls, operations, table_name, column_name, schema=None, **kw
    ):
        """Handle the different situations arising from dropping geospatial column from a DB."""
        op = cls(table_name, column_name, schema=schema, **kw)
        return operations.invoke(op)

    def reverse(self):
        """Used to autogenerate the downgrade function."""
        return AddGeospatialColumnOp.from_column_and_tablename(
            self.schema, self.table_name, self.column
        )

    @classmethod
    def batch_drop_geospatial_column(cls, operations, column_name, **kw):
        """Issue a "drop column" instruction using the current batch migration context."""
        op = cls(
            operations.impl.table_name,
            column_name,
            schema=operations.impl.schema,
            **kw,
        )
        return operations.invoke(op)


@Operations.implementation_for(AddGeospatialColumnOp)
def add_geospatial_column(operations, operation):
    """Handle the actual column addition according to the dialect backend.

    Args:
        operations: Operations object from alembic base, defining high level migration operations.
        operation: AddGeospatialColumnOp call, with attributes for table_name, column_name,
            column_type, and optional keywords.
    """

    table_name = operation.table_name
    column_name = operation.column.name

    dialect = operations.get_bind().dialect

    if dialect.name == "sqlite":
        if isinstance(operation.column, TypeDecorator):
            # Will be either geoalchemy2.types.Geography or geoalchemy2.types.Geometry, if using a
            # custom type
            geospatial_core_type = operation.column.type.load_dialect_impl(dialect)
        else:
            geospatial_core_type = operation.column.type
        operations.execute(
            func.AddGeometryColumn(
                table_name,
                column_name,
                geospatial_core_type.srid,
                geospatial_core_type.geometry_type,
                geospatial_core_type.dimension,
                not geospatial_core_type.nullable,
            )
        )
    elif "postgresql" in dialect.name:
        operations.impl.add_column(
            table_name,
            operation.column,
            schema=operation.schema,
        )


@Operations.implementation_for(DropGeospatialColumnOp)
def drop_geospatial_column(operations, operation):
    """Handle the actual column removal according to the dialect backend.

    Args:
        operations: Operations object from alembic base, defining high level migration operations.
        operation: AddGeospatialColumnOp call, with attributes for table_name, column_name,
            column_type, and optional keywords.
    """

    table_name = operation.table_name
    column = operation.to_column(operations.migration_context)

    dialect = operations.get_bind().dialect

    if dialect.name == "sqlite":
        _SPATIAL_TABLES.add(table_name)
    operations.impl.drop_column(
        table_name, column, schema=operation.schema, **operation.kw
    )


@compiles(RenameTable, "sqlite")
def visit_rename_geospatial_table(element, compiler, **kw):
    """Specific compilation rule to rename spatial tables with SQLite dialect."""
    table_is_spatial = element.table_name in _SPATIAL_TABLES
    new_table_is_spatial = element.new_table_name in _SPATIAL_TABLES

    if table_is_spatial or new_table_is_spatial:
        # Here we suppose there is only one DB attached to the current engine, so the prefix
        # is set to NULL
        return "SELECT RenameTable(NULL, '%s', '%s')" % (
            format_table_name(compiler, element.table_name, element.schema),
            format_table_name(compiler, element.new_table_name, element.schema),
        )
    else:
        return visit_rename_table(element, compiler, **kw)


@compiles(DropTable, "sqlite")
def visit_drop_geospatial_table(element, compiler, **kw):
    """Specific compilation rule to drop spatial tables with SQLite dialect."""
    table_is_spatial = element.element.name in _SPATIAL_TABLES

    if table_is_spatial:
        # Here we suppose there is only one DB attached to the current engine
        return "SELECT DropTable(NULL, '%s')" % (
            format_table_name(compiler, element.element.name, None),
        )
    else:
        return compiler.visit_drop_table(element, **kw)


@renderers.dispatch_for(AddGeospatialColumnOp)
def render_add_geo_column(autogen_context, op):
    """Render the add_geospatial_column operation in migration script."""
    col_render = _add_column(autogen_context, op)
    return col_render.replace(".add_column(", ".add_geospatial_column(")


@renderers.dispatch_for(DropGeospatialColumnOp)
def render_drop_geo_column(autogen_context, op):
    """Render the drop_geospatial_column operation in migration script."""
    col_render = _drop_column(autogen_context, op)
    return col_render.replace(".drop_column(", ".drop_geospatial_column(")


@writer.rewrites(ops.AddColumnOp)
def add_geo_column(context, revision, op):
    """Replace the default AddColumnOp by a geospatial-specific one."""
    col_type = op.column.type
    if isinstance(col_type, TypeDecorator):
        dialect = context.bind.dialect
        col_type = col_type.load_dialect_impl(dialect)
    if isinstance(col_type, (Geometry, Geography, Raster)):
        op.column.type.spatial_index = False
        op.column.type._spatial_index_reflected = None
        new_op = AddGeospatialColumnOp(op.table_name, op.column, op.schema)
    else:
        new_op = op
    return new_op


@writer.rewrites(ops.DropColumnOp)
def drop_geo_column(context, revision, op):
    """Replace the default DropColumnOp by a geospatial-specific one."""
    col_type = op.to_column().type
    if isinstance(col_type, TypeDecorator):
        dialect = context.bind.dialect
        col_type = col_type.load_dialect_impl(dialect)
    if isinstance(col_type, (Geometry, Geography, Raster)):
        new_op = DropGeospatialColumnOp(op.table_name, op.column_name, op.schema)
    else:
        new_op = op
    return new_op


@Operations.register_operation("create_geospatial_table")
class CreateGeospatialTableOp(ops.CreateTableOp):
    """
    Create a Geospatial Table in an Alembic migration context. This methodology originates from:
    https://alembic.sqlalchemy.org/en/latest/api/operations.html#operation-plugins
    """

    @classmethod
    def create_geospatial_table(cls, operations, table_name, *columns, **kw):
        """Handle the different situations arising from creating geospatial table to a DB."""
        op = cls(table_name, columns, **kw)
        return operations.invoke(op)

    def reverse(self):
        """Used to autogenerate the downgrade function."""
        return DropGeospatialColumnOp.from_table(
            self.to_table(),
            _namespace_metadata=self._namespace_metadata,
        )

    @classmethod
    def from_table(cls, table, _namespace_metadata=None):
        obj = super().from_table(table, _namespace_metadata)
        return obj

    def to_table(self, migration_context=None):
        table = super().to_table(migration_context)

        # Set spatial_index attribute to False so the indexes are created explicitly
        for col in table.columns:
            try:
                if col.type.spatial_index:
                    col.type.spatial_index = False
            except AttributeError:
                pass
        return table


@Operations.register_operation("drop_geospatial_table")
class DropGeospatialTableOp(ops.DropTableOp):
    @classmethod
    def drop_geospatial_table(cls, operations, table_name, schema=None, **kw):
        """Handle the different situations arising from dropping geospatial table from a DB."""

        op = cls(table_name, schema=schema, table_kw=kw)
        return operations.invoke(op)

    def reverse(self):
        """Used to autogenerate the downgrade function."""
        return CreateGeospatialTableOp.from_table(
            self.to_table(),
            _namespace_metadata=self._namespace_metadata,
        )

    @classmethod
    def from_table(cls, table, _namespace_metadata=None):
        obj = super().from_table(table, _namespace_metadata)
        return obj

    def to_table(self, migration_context=None):
        table = super().to_table(migration_context)
        return table


@Operations.implementation_for(CreateGeospatialTableOp)
def create_geospatial_table(operations, operation):
    """Handle the actual table creation according to the dialect backend.

    Args:
        operations: Operations object from alembic base, defining high level migration operations.
        operation: CreateGeospatialTableOp call, with attributes for table_name, column_name,
            column_type, and optional keywords.
    """
    table_name = operation.table_name
    bind = operations.get_bind()

    # For now the default events defined in geoalchemy2 are enought to handle table creation
    operations.create_table(table_name, *operation.columns, **operation.kw)

    if bind.dialect.name == "sqlite":
        _SPATIAL_TABLES.add(table_name)


@Operations.implementation_for(DropGeospatialTableOp)
def drop_geospatial_table(operations, operation):
    """Handle the actual table removal according to the dialect backend.

    Args:
        operations: Operations object from alembic base, defining high level migration operations.
        operation: DropGeospatialTableOp call, with attributes for table_name, column_name,
            column_type, and optional keywords.
    """
    table_name = operation.table_name
    bind = operations.get_bind()
    dialect = bind.dialect

    if dialect.name == "sqlite":
        _SPATIAL_TABLES.add(table_name)
    operations.drop_table(table_name, operation.schema, **operation.table_kw)


@renderers.dispatch_for(CreateGeospatialTableOp)
def render_create_geo_table(autogen_context, op):
    """Render the create_geospatial_table operation in migration script."""
    table_render = _add_table(autogen_context, op)
    return table_render.replace(".create_table(", ".create_geospatial_table(")


@renderers.dispatch_for(DropGeospatialTableOp)
def render_drop_geo_table(autogen_context, op):
    """Render the drop_geospatial_table operation in migration script."""
    table_render = _drop_table(autogen_context, op)
    return table_render.replace(".drop_table(", ".drop_geospatial_table(")


@writer.rewrites(ops.CreateTableOp)
def create_geo_table(context, revision, op):
    """Replace the default CreateTableOp by a geospatial-specific one."""
    dialect = context.bind.dialect
    gis_cols = _get_gis_cols(
        op, (Geometry, Geography, Raster), dialect, check_col_management=False
    )

    if gis_cols:
        new_op = CreateGeospatialTableOp(op.table_name, op.columns, op.schema)
    else:
        new_op = op

    return new_op


@writer.rewrites(ops.DropTableOp)
def drop_geo_table(context, revision, op):
    """Replace the default DropTableOp by a geospatial-specific one."""
    dialect = context.bind.dialect
    table = op.to_table()
    gis_cols = _get_gis_cols(
        table, (Geometry, Geography, Raster), dialect, check_col_management=False
    )

    if gis_cols:
        new_op = DropGeospatialTableOp(op.table_name, op.schema)
    else:
        new_op = op

    return new_op


@Operations.register_operation("create_geospatial_index")
@BatchOperations.register_operation(
    "create_geospatial_index", "batch_create_geospatial_index"
)
class CreateGeospatialIndexOp(ops.CreateIndexOp):
    @classmethod
    def create_geospatial_index(
        cls,
        operations,
        index_name,
        table_name,
        columns,
        schema=None,
        unique=False,
        **kw,
    ):
        """Handle the different situations arising from creating geospatial index into a DB."""
        op = cls(index_name, table_name, columns, schema=schema, unique=unique, **kw)
        return operations.invoke(op)

    def reverse(self):
        """Used to autogenerate the downgrade function."""
        return DropGeospatialIndexOp(
            self.index_name,
            self.table_name,
            column_name=self.columns[0].name,
            schema=self.schema,
        )

    @classmethod
    def batch_create_geospatial_index(
        cls,
        operations,
        index_name,
        columns,
        **kw,
    ):
        """Issue a "create index" instruction using the current batch migration context."""
        op = cls(
            index_name,
            operations.impl.table_name,
            columns,
            schema=operations.impl.schema,
            **kw,
        )
        return operations.invoke(op)


@Operations.register_operation("drop_geospatial_index")
@BatchOperations.register_operation(
    "drop_geospatial_index", "batch_drop_geospatial_index"
)
class DropGeospatialIndexOp(ops.DropIndexOp):
    def __init__(self, *args, column_name, **kwargs):
        super().__init__(*args, **kwargs)
        self.column_name = column_name

    @classmethod
    def drop_geospatial_index(
        cls,
        operations,
        index_name,
        table_name,
        column_name,
        schema=None,
        unique=False,
        **kw,
    ):
        """Handle the different situations arising from dropping geospatial index from a DB."""
        op = cls(
            index_name,
            table_name,
            column_name=column_name,
            schema=schema,
            unique=unique,
            **kw,
        )
        return operations.invoke(op)

    def reverse(self):
        """Used to autogenerate the downgrade function."""
        return CreateGeospatialIndexOp(
            self.index_name,
            self.table_name,
            column_name=self.column_name,
            schema=self.schema,
            _reverse=self,
            **self.kw,
        )

    @classmethod
    def from_index(cls, index):
        assert index.table is not None
        assert len(index.columns) == 1, "A spatial index must be set on one column only"
        return cls(
            index.name,
            index.table.name,
            column_name=index.columns[0].name,
            schema=index.table.schema,
            _reverse=CreateGeospatialIndexOp.from_index(index),
            **index.kwargs,
        )

    @classmethod
    def batch_drop_geospatial_index(cls, operations, index_name, **kw):
        """Issue a "drop index" instruction using the current batch migration context."""
        op = cls(
            index_name,
            table_name=operations.impl.table_name,
            schema=operations.impl.schema,
            **kw,
        )
        return operations.invoke(op)


@Operations.implementation_for(CreateGeospatialIndexOp)
def create_geospatial_index(operations, operation):
    """Handle the actual index creation according to the dialect backend.

    Args:
        operations: Operations object from alembic base, defining high level migration operations.
        operation: CreateGeospatialIndexOp call, with attributes for table_name, column_name,
            column_type, and optional keywords.
    """
    bind = operations.get_bind()
    dialect = bind.dialect

    if dialect.name == "sqlite":
        assert (
            len(operation.columns) == 1
        ), "A spatial index must be set on one column only"
        operations.execute(
            func.CreateSpatialIndex(operation.table_name, operation.columns[0])
        )
    else:
        idx = operation.to_index(operations.migration_context)
        operations.impl.create_index(idx)


@Operations.implementation_for(DropGeospatialIndexOp)
def drop_geospatial_index(operations, operation):
    """Handle the actual index drop according to the dialect backend.

    Args:
        operations: Operations object from alembic base, defining high level migration operations.
        operation: DropGeospatialIndexOp call, with attributes for table_name, column_name,
            column_type, and optional keywords.
    """
    bind = operations.get_bind()
    dialect = bind.dialect

    if dialect.name == "sqlite":
        operations.execute(
            func.DisableSpatialIndex(operation.table_name, operation.column_name)
        )
    else:
        operations.impl.drop_index(operation.to_index(operations.migration_context))


@renderers.dispatch_for(CreateGeospatialIndexOp)
def render_create_geo_index(autogen_context, op):
    """Render the create_geospatial_index operation in migration script."""
    idx_render = _add_index(autogen_context, op)
    return idx_render.replace(".create_index(", ".create_geospatial_index(")


@renderers.dispatch_for(DropGeospatialIndexOp)
def render_drop_geo_index(autogen_context, op):
    """Render the drop_geospatial_index operation in migration script."""
    idx_render = _drop_index(autogen_context, op)

    # Replace function name
    text = idx_render.replace(".drop_index(", ".drop_geospatial_index(")

    # Add column name as keyword argument
    text = text[:-1] + ", column_name='%s')" % (op.column_name,)

    return text


@writer.rewrites(ops.CreateIndexOp)
def create_geo_index(context, revision, op):
    """Replace the default CreateIndexOp by a geospatial-specific one."""
    dialect = context.bind.dialect

    if len(op.columns) == 1:
        col = op.columns[0]
        if isinstance(col, Column) and _check_spatial_type(
            col.type, (Geometry, Geography, Raster), dialect
        ):
            # Fix index properties
            op.kw["postgresql_using"] = op.kw.get("postgresql_using", "gist")
            if col.type.use_N_D_index:
                postgresql_ops = {col.name: "gist_geometry_ops_nd"}
            else:
                postgresql_ops = {}
            op.kw["postgresql_ops"] = op.kw.get("postgresql_ops", postgresql_ops)

            return CreateGeospatialIndexOp(
                op.index_name, op.table_name, op.columns, op.schema, op.unique, **op.kw
            )

    return op


@writer.rewrites(ops.DropIndexOp)
def drop_geo_index(context, revision, op):
    """Replace the default DropIndexOp by a geospatial-specific one."""
    dialect = context.bind.dialect
    idx = op.to_index()

    if len(idx.columns) == 1:
        col = idx.columns[0]
        if isinstance(col, Column) and _check_spatial_type(
            col.type, (Geometry, Geography, Raster), dialect
        ):
            return DropGeospatialIndexOp(
                op.index_name,
                op.table_name,
                column_name=col.name,
                schema=op.schema,
                **op.kw,
            )

    return op



"""

This module defines a ``Comparator`` class for use with geometry and geography
objects. This is where spatial operators, like ``&&``, ``&<``, are defined.
Spatial operators very often apply to the bounding boxes of geometries. For
example, ``geom1 && geom2`` indicates if geom1's bounding box intersects
geom2's.

Examples
--------

Select the objects whose bounding boxes are to the left of the
bounding box of ``POLYGON((-5 45,5 45,5 -45,-5 -45,-5 45))``::

    select([table]).where(table.c.geom.to_left(
        'POLYGON((-5 45,5 45,5 -45,-5 -45,-5 45))'))

The ``<<`` and ``>>`` operators are a bit specific, because they have
corresponding Python operator (``__lshift__`` and ``__rshift__``). The
above ``SELECT`` expression can thus be rewritten like this::

    select([table]).where(
        table.c.geom << 'POLYGON((-5 45,5 45,5 -45,-5 -45,-5 45))')

Operators can also be used when using the ORM. For example::

    Session.query(Cls).filter(
        Cls.geom << 'POLYGON((-5 45,5 45,5 -45,-5 -45,-5 45))')

Now some other examples with the ``<#>`` operator.

Select the ten objects that are the closest to ``POINT(0 0)`` (typical
closed neighbors problem)::

    select([table]).order_by(table.c.geom.distance_box('POINT(0 0)')).limit(10)

Using the ORM::

    Session.query(Cls).order_by(Cls.geom.distance_box('POINT(0 0)')).limit(10)

Reference
---------
"""

from sqlalchemy import types as sqltypes
from sqlalchemy.dialects.postgresql import DOUBLE_PRECISION
from sqlalchemy.sql import operators
from sqlalchemy.types import UserDefinedType

try:
    from sqlalchemy.sql.functions import _FunctionGenerator
except ImportError:  # SQLA < 0.9  # pragma: no cover
    from sqlalchemy.sql.expression import _FunctionGenerator


INTERSECTS = operators.custom_op('&&')
INTERSECTS_ND = operators.custom_op('&&&')
OVERLAPS_OR_TO_LEFT = operators.custom_op('&<')
OVERLAPS_OR_TO_RIGHT = operators.custom_op('&>')
OVERLAPS_OR_BELOW = operators.custom_op('&<|')
TO_LEFT = operators.custom_op('<<')
BELOW = operators.custom_op('<<|')
TO_RIGHT = operators.custom_op('>>')
CONTAINED = operators.custom_op('@')
OVERLAPS_OR_ABOVE = operators.custom_op('|&>')
ABOVE = operators.custom_op('|>>')
CONTAINS = operators.custom_op('~')
SAME = operators.custom_op('~=')
DISTANCE_CENTROID = operators.custom_op('<->')
DISTANCE_BOX = operators.custom_op('<#>')


class BaseComparator(UserDefinedType.Comparator):
    """
    A custom comparator base class. It adds the ability to call spatial
    functions on columns that use this kind of comparator. It also defines
    functions that map to operators supported by ``Geometry``, ``Geography``
    and ``Raster`` columns.

    This comparator is used by the :class:`geoalchemy2.types.Raster`.
    """

    key = None

    def __getattr__(self, name):

        # Function names that don't start with "ST_" are rejected.
        # This is not to mess up with SQLAlchemy's use of
        # hasattr/getattr on Column objects.

        if not name.lower().startswith('st_'):
            raise AttributeError

        # We create our own _FunctionGenerator here, and use it in place of
        # SQLAlchemy's "func" object. This is to be able to "bind" the
        # function to the SQL expression. See also GenericFunction.

        func_ = _FunctionGenerator(expr=self.expr)
        return getattr(func_, name)

    def intersects(self, other):
        """
        The ``&&`` operator. A's BBOX intersects B's.
        """
        return self.operate(INTERSECTS, other, result_type=sqltypes.Boolean)

    def overlaps_or_to_left(self, other):
        """
        The ``&<`` operator. A's BBOX overlaps or is to the left of B's.
        """
        return self.operate(OVERLAPS_OR_TO_LEFT, other,
                            result_type=sqltypes.Boolean)

    def overlaps_or_to_right(self, other):
        """
        The ``&>`` operator. A's BBOX overlaps or is to the right of B's.
        """
        return self.operate(OVERLAPS_OR_TO_RIGHT, other,
                            result_type=sqltypes.Boolean)


class Comparator(BaseComparator):
    """
    A custom comparator class. Used in :class:`geoalchemy2.types.Geometry`
    and :class:`geoalchemy2.types.Geography`.

    This is where spatial operators like ``<<`` and ``<->`` are defined.
    """

    def overlaps_or_below(self, other):
        """
        The ``&<|`` operator. A's BBOX overlaps or is below B's.
        """
        return self.operate(OVERLAPS_OR_BELOW, other,
                            result_type=sqltypes.Boolean)

    def to_left(self, other):
        """
        The ``<<`` operator. A's BBOX is strictly to the left of B's.
        """
        return self.operate(TO_LEFT, other, result_type=sqltypes.Boolean)

    def __lshift__(self, other):
        """
        The ``<<`` operator. A's BBOX is strictly to the left of B's.
        Same as ``to_left``, so::

            table.c.geom << 'POINT(1 2)'

        is the same as::

            table.c.geom.to_left('POINT(1 2)')
        """
        return self.to_left(other)

    def below(self, other):
        """
        The ``<<|`` operator. A's BBOX is strictly below B's.
        """
        return self.operate(BELOW, other, result_type=sqltypes.Boolean)

    def to_right(self, other):
        """
        The ``>>`` operator. A's BBOX is strictly to the right of B's.
        """
        return self.operate(TO_RIGHT, other, result_type=sqltypes.Boolean)

    def __rshift__(self, other):
        """
        The ``>>`` operator. A's BBOX is strictly to the left of B's.
        Same as `to_`right``, so::

            table.c.geom >> 'POINT(1 2)'

        is the same as::

            table.c.geom.to_right('POINT(1 2)')
        """
        return self.to_right(other)

    def contained(self, other):
        """
        The ``@`` operator. A's BBOX is contained by B's.
        """
        return self.operate(CONTAINED, other, result_type=sqltypes.Boolean)

    def overlaps_or_above(self, other):
        """
        The ``|&>`` operator. A's BBOX overlaps or is above B's.
        """
        return self.operate(OVERLAPS_OR_ABOVE, other,
                            result_type=sqltypes.Boolean)

    def above(self, other):
        """
        The ``|>>`` operator. A's BBOX is strictly above B's.
        """
        return self.operate(ABOVE, other, result_type=sqltypes.Boolean)

    def contains(self, other, **kw):
        """
        The ``~`` operator. A's BBOX contains B's.
        """
        return self.operate(CONTAINS, other, result_type=sqltypes.Boolean)

    def same(self, other):
        """
        The ``~=`` operator. A's BBOX is the same as B's.
        """
        return self.operate(SAME, other, result_type=sqltypes.Boolean)

    def distance_centroid(self, other):
        """
        The ``<->`` operator. The distance between two points.
        """
        return self.operate(DISTANCE_CENTROID, other,
                            result_type=DOUBLE_PRECISION)

    def distance_box(self, other):
        """
        The ``<#>`` operator. The distance between bounding box of two
        geometries.
        """
        return self.operate(DISTANCE_BOX, other, result_type=DOUBLE_PRECISION)

    def intersects_nd(self, other):
        """
        The ``&&&`` operator returns TRUE if the n-D bounding box of geometry A
        intersects the n-D bounding box of geometry B.
        """
        return self.operate(INTERSECTS_ND, other, result_type=sqltypes.Boolean)



"""
Function translation for specific dialect
=========================================

Some functions have different names depending on the dialect. But sometimes one function in one
dialect can be mapped to several other functions in another dialect, depending on the arguments
passed. For example, the `ST_Buffer` function in PostgreSQL can translate into 2 functions in
SQLite:

1. if the buffer is two-sided (symmetric), the PostgreSQL function::

    ST_Buffer(the_table.geom, 10)

  should become in SQLite::

    Buffer(the_table.geom, 10)

2. if the buffer is one-sided, the PostgreSQL function::

    ST_Buffer(the_table.geom, 10, 'side=right')

  should become in SQLite::

    SingleSidedBuffer(the_table.geom, 10, 0)

This case is much more complicated than just mapping a function name and we show here how to deal
with it.

This example uses SQLAlchemy core queries.
"""
from sqlalchemy import MetaData
from sqlalchemy import func
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.sql.expression import BindParameter

from geoalchemy2 import WKTElement
from geoalchemy2 import functions

# Tests imports
from tests import format_wkt
from tests import select

metadata = MetaData()
Base = declarative_base(metadata=metadata)


def _compile_buffer_default(element, compiler, **kw):
    """Compile the element in the default case (no specific dialect).

    This function should not be needed for SQLAlchemy >= 1.1.
    """
    return '{}({})'.format('ST_Buffer', compiler.process(element.clauses, **kw))


def _compile_buffer_sqlite(element, compiler, **kw):
    """Compile the element for the SQLite dialect."""
    # Get the side parameters
    compiled = compiler.process(element.clauses, **kw)
    side_params = [
        i for i in element.clauses
        if isinstance(i, BindParameter) and 'side' in str(i.value)
    ]

    if side_params:
        side_param = side_params[0]
        if 'right' in side_param.value:
            # If the given side is 'right', we translate the value into 0 and switch to the sided
            # function
            side_param.value = 0
            element.identifier = 'SingleSidedBuffer'
        elif 'left' in side_param.value:
            # If the given side is 'left', we translate the value into 1 and switch to the sided
            # function
            side_param.value = 1
            element.identifier = 'SingleSidedBuffer'

    if element.identifier == 'ST_Buffer':
        # If the identifier is still the default ST_Buffer we switch to the SpatiaLite function
        element.identifier = 'Buffer'

    # If there is no side parameter or if the side value is 'both', we use the default function
    return '{}({})'.format(element.identifier, compiled)


# Register the specific compilation rules
compiles(functions.ST_Buffer)(_compile_buffer_default)
compiles(functions.ST_Buffer, 'sqlite')(_compile_buffer_sqlite)


def test_specific_compilation(conn):
    # Build a query with a sided buffer
    query = select([
        func.ST_AsText(
            func.ST_Buffer(WKTElement('LINESTRING(0 0, 1 0)', srid=4326), 1, 'side=left')
        )
    ])

    # Check the compiled query: the sided buffer should appear only in the SQLite query
    compiled_query = str(query.compile(dialect=conn.dialect))
    if conn.dialect.name == 'sqlite':
        assert 'SingleSidedBuffer' in compiled_query
        assert 'ST_Buffer' not in compiled_query
    else:
        assert 'SingleSidedBuffer' not in compiled_query
        assert 'ST_Buffer' in compiled_query

    # Check the actual result of the query
    res = conn.execute(query).scalar()
    assert format_wkt(res) == 'POLYGON((1 0,0 0,0 1,1 1,1 0))'

    # Build a query with symmetric buffer to check nothing was broken
    query = select([
        func.ST_AsText(
            func.ST_Buffer(WKTElement('LINESTRING(0 0, 1 0)', srid=4326), 1)
        )
    ])

    # Check the compiled query: the sided buffer should never appear in the query
    compiled_query = str(query.compile(dialect=conn.dialect))
    assert 'SingleSidedBuffer' not in compiled_query
    if conn.dialect.name == 'sqlite':
        assert 'ST_Buffer' not in compiled_query
        assert 'Buffer' in compiled_query
    else:
        assert 'ST_Buffer' in compiled_query

    # Check the actual result of the query
    res = conn.execute(query).scalar()
    assert format_wkt(res) != 'POLYGON((1 0,0 0,0 1,1 1,1 0))'
    assert format_wkt(res).startswith('POLYGON((1 1,1')



"""Exceptions used with GeoAlchemy2."""


class GeoAlchemyError(Exception):
    """Generic error class."""


class ArgumentError(GeoAlchemyError):
    """Raised when an invalid or conflicting function argument is supplied."""
